import os
import logging
import shutil

from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple

from dataclasses import dataclass

from pathlib import Path

import numpy as np
import tensorflow as tf
import transformers

from tensorflow.keras.optimizers.experimental import AdamW
from tensorflow.keras.metrics import (
    Mean,
    Precision,
    Recall,
)

from transformers import (
     AutoConfig,
     AutoTokenizer,
     PretrainedConfig,
     PreTrainedTokenizerFast,
     TFPreTrainedModel,
     TFAutoModelForMaskedLM,
)

from tokenizers.pre_tokenizers import Whitespace

from nlpipes.trainers.trainers import Trainer
from nlpipes.losses.losses import WCELoss
from nlpipes.metrics.confusion import F1_Score

from nlpipes.data.data_loaders import DataLoader
from nlpipes.data.data_selectors import DataSelector
from nlpipes.data.data_augmentors import VocabAugmentor
from nlpipes.data.data_processors import MultiLabelProcessor
from nlpipes.data.data_cleaners import clean

from nlpipes.data.data_utils import (
     create_multilabel_examples,
     split_examples,
     generate_batches,
     chunk,
     convert_onehot_to_ids,
     convert_ids_to_onehot,
)

from nlpipes.data.data_types import (
     Example,
     Features,
     Prediction,
     Classification,
)

from nlpipes.callbacks.callbacks import (
     TrainingStep,
     History,
     LossHistory,
     FBetaScoreHistory,
     F1ScoreHistory,
     PrecisionHistory,
     RecallHistory,
     ModelCheckpoint,
     CSVLogger,
     ProgbarLogger,
     EarlyStopping,
     TimedStopping,
)

transformers.logging.set_verbosity_error()

logger = logging.getLogger('__name__')


@dataclass
class PipelineForMultiLabelSequenceClassification():
    """ The pipeline for aspect category classification.
    
    Args
    ----------
    model(TFPreTrainedModel):
       The model used for the task`.
    
    tokenizer(PreTrainedTokenizerFast):
       The rust tokenizer used to convert the input
       text into encoded vectors.
    
    config(PretrainedConfig): 
       The model configuration file.
    """
    
    model: TFPreTrainedModel
    tokenizer: PreTrainedTokenizerFast
    config: PretrainedConfig
    
    def train(self,
              X: List[str] = None, 
              Y: List[str] =  None,
              train_frac: float = 0.8,
              test_frac: float = 0.2,
              class_weights: List[float] = None, 
              epochs: int = 3,
              batch_size: int = 16,
              thresholds: float = 0.5,
              max_length: int = 128,
              lr: float = 2e-5,
              beta_1: float = 0.9,
              beta_2: float = 0.999,
              patience: int = 99999,
              min_delta: float = 0.01,
              time_limit: int = 86400,
              random_seed: int = 69,
              keep_checkpoints: bool = False,
              verbose: bool = True,
              checkpoints_dir: str = './model/checkpoints',
              logs_dir: str = './model',
             ) -> Tuple[History]:
        """ Train a model for multi-label classification.
        Most of the training process is implemented through 
        callbacks functions. Even the actual training step 
        (calculating the gradient and updating the weights 
        using an optimizer) is implemented as callbacks and
        is not part of the Trainer itself. This provide a 
        modular architecture that allows new ideas to be 
        implemented without having to change and increase
        the complexity of the trainer. """
        
        examples = create_multilabel_examples(
            texts=X, 
            labels=Y,
            config=self.config,
        )
        
        train_examples, test_examples = split_examples(
            examples,
            train_frac=train_frac,
            test_frac=test_frac,
            shuffle=True,
            random_seed=random_seed,
        )
        
        data_processor = MultiLabelProcessor(
            name='DataProcessor',
            tokenizer=self.tokenizer,
            config=self.config,
            max_length=max_length,
        )
        
        train_stream = DataLoader(
            name='TrainDataLoader',
            examples=train_examples,
            batch_size=batch_size,
            data_processor=data_processor,
            random_seed=random_seed,
        )
        
        test_stream = DataLoader(
            name='TestDataLoader',
            examples=test_examples,
            batch_size=batch_size,
            data_processor=data_processor,
            random_seed=random_seed,
        )
        
        training_step = TrainingStep(
            name='TrainingStep',
            model=self.model,
            loss_fn=WCELoss(class_weights),
            optimizer=AdamW(
                learning_rate=lr,
                beta_1=beta_1,
                beta_2=beta_2,
            ),
        )
        
        loss_history = LossHistory(
            name='Loss', 
            metric=Mean(),
            training_step=training_step,
            verbose=verbose,
        )
        
        f1_history = F1ScoreHistory(
            name='F1',
            training_step=training_step,
            metric=F1_Score(
                thresholds=thresholds, 
                top_k=1,
            ),
            mode='multilabel',
            verbose=verbose,
        )
        
        precision_history = PrecisionHistory(
            name='Precision',
            training_step=training_step,
            metric=Precision(
                thresholds, 
                top_k=1,
            ),
            mode='multilabel',
            verbose=verbose,
        )
        
        recall_history = RecallHistory(
            name='Recall',
            training_step=training_step,
            metric=Recall(
                thresholds, 
                top_k=1,
            ),
            mode='multilabel',
            verbose=verbose,
        )
        
        model_checkpoint = ModelCheckpoint(
            name='ModelCheckpoint',
            history=loss_history,
            model=self.model, 
            direction='minimize',
            checkpoints_dir=checkpoints_dir,
        )
        
        early_stopping = EarlyStopping(
            name='EarlyStopping',
            history=loss_history,
            min_delta=min_delta,
            direction='minimize',
            patience=patience,
        )
        
        timed_stopping = TimedStopping(
            name='TimedStopping',
            time_limit=time_limit,
        )
        
        csv_logger = CSVLogger(
            name='CSVLogger',
            logs_dir=logs_dir,
        )
        
        progbar_logger = ProgbarLogger(
            name='ProgbarLogger',
            batch_size=batch_size,
            history=[
                loss_history,
                f1_history,
                precision_history,
                recall_history,
            ],
            num_samples=sum([
                train_stream.num_examples(),
                test_stream.num_examples(),
            ])
        )
        
        callbacks = [
            training_step,
            loss_history,
            f1_history,
            precision_history,
            recall_history,
            model_checkpoint,
            csv_logger,
            progbar_logger,
            early_stopping,
            timed_stopping,
        ]
        
        trainer = Trainer(callbacks)
        
        trainer.fit(
            train_stream=train_stream,
            test_stream=test_stream,
            epochs=epochs,
        )
        
        if keep_checkpoints == False:
            shutil.rmtree(checkpoints_dir)
            
        return (
            loss_history,
            f1_history, 
            precision_history,
            recall_history,
        )
    
    def evaluate(self,
                 X: List[str],
                 Y: List[str],
                 metric: tf.metrics.Metric,
                 batch_size: int = 32,
                 thresholds: float = 0.5,
                ) -> tf.metrics.Metric:
        """ Evaluate the model performance on new data
        according to the user-defined evaluation metric """
        
        for texts, labels in generate_batches(X, Y, batch_size):
            examples = self.ingest(texts, labels)
            features = self.transform(examples)
            predictions = self.predict(features, thresholds)
            
            y_true = [e.label for e in examples]
            y_pred = [p.label for p in predictions]
              
            y_true = convert_ids_to_onehot(
                ids=y_true,
                depth=self.config.num_labels,
                dtype=tf.float32
            )
            
            metric.update_state(y_true, y_pred)
            
        return metric.result()
    
    def classify(
        self,
        texts: List[str],
        thresholds: float = 0.5,
        chunk_text: bool = False,
    ) -> List[Classification]:
        """ Get predictions on new input texts """
        
        examples = self.ingest(texts)
        features = self.transform(examples)
        predictions = self.predict(features, thresholds)
        classifications = self.postprocess(predictions, examples, thresholds)
        
        return classifications
    
    def ingest(
        self,
        texts: List[str],
        labels: Optional[str] = None,
        chunk_text: Optional[bool] = False,
    ) -> List[Example]:
        """ Convert each raw text into examples. 
        Raw texts can eventually be chunk into spans """
        
        if labels:
            labels_id = [
                [
                    self.config.label2id[label] 
                    for label in sub_labels
                ]
                for sub_labels in labels
            ]
            examples = [
                Example(text, label) 
                for text, label
                in zip(texts, labels_id)
            ]
        else:
            texts = chunk(texts) if chunk_text else texts
            examples = [Example(text) for text in texts]
            
        return examples
    
    def transform(self, examples: Iterable[Example]) -> Features: 
        """ Encode human-readable examples into encoded
        features that will be used as input by the model. """
        
        texts = [e.text for e in examples]
        max_length = self.config.max_position_embeddings
        
        inputs = self.tokenizer(
           texts,
           add_special_tokens=True,
           padding=True,
           truncation=True,
           max_length=max_length,
           is_split_into_words = False,
           return_attention_mask=True,
           return_token_type_ids=True,
           return_tensors='tf',
        )
        
        features = Features(
           input_ids=inputs['input_ids'],
           attention_mask=inputs['attention_mask'],
           token_type_ids=inputs['token_type_ids'],
        )
        
        return features
    
    def predict(
        self,
        features: Features, 
        thresholds: float=0.5,
    ) -> List[Prediction]:
        """ Predict the label indices where prediction score
        is over the user defined threshold and compute the loss
        value associated with each prediction. """
        
        outputs = self.model.call(
            input_ids=features.input_ids,
            attention_mask=features.attention_mask,
            token_type_ids=features.token_type_ids,
        )
        
        logits = outputs.logits
        confidences = tf.nn.sigmoid(logits)
        thresholds = tf.constant(thresholds)
        labels = tf.cast(
            tf.math.greater(confidences, thresholds), 
            tf.float32,
        )
        
        predictions = [
            Prediction(label, confidence)
            for label, confidence
            in zip(labels, confidences)
        ]
        
        return predictions
    
    def postprocess(
        self,
        predictions: Iterable[Prediction],
        examples: Iterable[Example],
        thresholds: float,
    ) -> List[Classification]:
        """ Post-process the predictions. We convert tensors 
        into more convenient python lists and alss convert the 
        predicted labels ids into their original name. """
        
        texts = [e.text for e in examples]
        id2label = self.config.id2label
        
        labels = [
            convert_onehot_to_ids(p.label) 
            for p in predictions
        ]
        
        labels = [
            [id2label[l.numpy()] for l in label]
            for label in labels
        ]
        
        confidences = [
            p.confidence.numpy()
            for p in predictions
        ]
        
        confidences = [
            c[c > thresholds] 
            for c in confidences
        ]
        
        confidences = [c.tolist() for c in confidences]
        
        classifications = [
            Classification(text, label, confidence)
            for text, label, confidence
            in zip(texts, labels, confidences)
        ]

        return classifications
    
    def save(self, save_dir: str):
        """ Save the trained model on disk """
        
        os.makedirs(save_dir, exist_ok=True)
        self.model.save_pretrained(save_dir)
        self.tokenizer.save_pretrained(save_dir)
        self.config.save_pretrained(save_dir)



from typing import Callable
from typing import List

from dataclasses import dataclass

import copy

import numpy as np
import tensorflow as tf

from transformers import (
    PretrainedConfig,
    PreTrainedTokenizer,
    AutoTokenizer,
)

from nlpipes.data.data_types import Example
from nlpipes.data.data_types import Features


""" The data processor classes.

The data processor encapsulates all the transformations
needed to encode input examples into features to be used
as input batch by the model. """


@dataclass
class MLMProcessor:
    """ 
    Data processor for Masked Language Modeling.
    By default, we assign a 15% probability for 
    each token of a sequence to be included in
    the mask. Among the tokens in the mask:
     - 80% are replaced with [MASK] token
     - 10% are replaced with some random token
     - 10% are kept unchanged.
    The special tokens i.e., [CLS], [SEP], [PAD] 
    are excluded from the mask, as we don’t want 
    to predict them during training.
    Inputs are dynamically padded to the maximum 
    length of a batch if they are not all of the 
    same length
    
    Args
    ----------

    tokenizer:
        The tokenizer used for encoding the data.

    config:
        The configuration of the model.
    """
    
    name: str = 'MaskedLanguageModelingProcessor'
    tokenizer: PreTrainedTokenizer = None
    config: PretrainedConfig = None
    max_length: int = 128
    mlm_prob: float = 0.15
    mask_prob: float = 0.80
    rand_prob: float = 0.10
    idem_prob: float = 0.10
    
    def __call__(self, 
                 examples: List[Example],
                ) -> Features:
        
        texts = [example.text for example in examples]
        
        # We first tokenize the input sequence
        inputs = self.tokenizer(
            texts,
            add_special_tokens=True,
            padding=True,
            truncation=True,
            max_length=self.max_length,
            is_split_into_words=False,
            return_attention_mask=True,
            return_token_type_ids=True,
            return_tensors='tf',
        )
        
        # The target label is simply the input_ids
        labels = tf.identity(inputs.input_ids)
        
        # Random selection of tokens to be masked. 
        mask, mask_details = self.create_mask(
            inputs=inputs.input_ids,
            excluded=self.tokenizer.all_special_ids,
            mlm_prob=self.mlm_prob,
            mask_prob=self.mask_prob,
            rand_prob=self.rand_prob,
            idem_prob=self.idem_prob,
        )
        
        # 80% of the time, replace with the [MASK] token
        inputs['input_ids'] = self.apply_mask(
            inputs=inputs.input_ids,
            mask=mask_details['masked'],
            mask_id=self.tokenizer.mask_token_id,
        )
        
        # 10% of the time, replace with random token
        vocab_size = len(self.tokenizer)
        random_token_ids = tf.random.uniform(
            shape=inputs.input_ids.shape, 
            maxval=vocab_size,
            dtype=inputs.input_ids.dtype,
        )
        
        inputs['input_ids'] = self.apply_mask(
            inputs=inputs.input_ids,
            mask=mask_details['random'],
            mask_id=random_token_ids,
        )
        
        # Unchanged The rest of the time (10% of the time).
        
        # Only the `masked` tokens are predicted during training.
        labels = self.apply_mask(
            inputs=labels,
            mask=~mask,
            mask_id=-100,
        )
        
        inputs['label'] = tf.one_hot(
            [label for label in labels],
            depth=vocab_size
        )
        
        input_features = Features(
            input_ids=inputs['input_ids'],
            attention_mask=inputs['attention_mask'],
            token_type_ids=inputs['token_type_ids'],
            label=inputs['label'],
        )
        
        return input_features
    
    def create_mask(
        self,
        inputs: tf.Tensor,
        mlm_prob: float,
        mask_prob: float,
        rand_prob: float,
        idem_prob: float,
        excluded: tf.Tensor,  
    ) -> tf.Tensor:
        """ Create the mask used for the masked 
        language modeling."""
        
        uniform = tf.random.uniform(inputs.shape, maxval=1)
        
        mask_treshold = mlm_prob * mask_prob
        rand_treshold = mlm_prob * (mask_prob + rand_prob)

        mask = tf.math.logical_and(
            uniform <= mlm_prob, 
            ~np.isin(inputs, excluded),
        )
        
        masked = tf.math.logical_and(
            uniform <= mask_treshold, 
            ~np.isin(inputs, excluded),
        )
        
        random = tf.math.logical_and(
            uniform > mask_treshold,
            uniform <= rand_treshold,
            ~np.isin(inputs, excluded),
        )
        
        mask_details = {}
        mask_details['masked'] = masked
        mask_details['random'] = random
        
        return mask, mask_details
    
    def apply_mask(
        self,
        inputs: tf.Tensor,
        mask: tf.Tensor,
        mask_id: int,
    ) -> tf.Tensor:
        """ Apply the mask to the given inputs """
        
        input_masked = tf.where(
            condition=mask,
            x=mask_id,
            y=inputs,
        )
        
        return input_masked


@dataclass
class SingleLabelProcessor:
    """
    Data Processor for single-label sequence classification.
    Inputs are dynamically padded to the maximum 
    length of a batch if they are not all of the 
    same length.
    
    Args
    ----------

    tokenizer (transformers.PreTrainedTokenizer):
        The tokenizer used for encoding the data.

    config (transformers.PreTrainedCOnfig):
        The configuration of the model.
    """
    
    name: str = 'SingleLabelClassificationProcessor'
    tokenizer: PreTrainedTokenizer = None
    config: PretrainedConfig = None
    max_length: int = None
    
    def __call__(self, 
                 examples: List[Example],
                ) -> Features:
        
        texts = [example.text for example in examples]
        
        inputs = self.tokenizer(
            texts,
            add_special_tokens=True,
            padding=True,
            truncation=True,
            max_length=self.max_length,
            is_split_into_words=False,
            return_attention_mask=True,
            return_token_type_ids=True,
            return_tensors='tf',
        )
        
        inputs['label'] = tf.one_hot(
            [example.label for example in examples],
            depth=self.config.num_labels
        )
        
        input_features = Features(
            input_ids=inputs['input_ids'],
            attention_mask=inputs['attention_mask'],
            token_type_ids=inputs['token_type_ids'],
            label=inputs['label'],
        )
        
        return input_features


@dataclass(frozen=True)
class MultiLabelProcessor:
    """
    Data Processor for multi-label sequence classification.
    Inputs are dynamically padded to the maximum length of a
    batch if they are not all of the same length.
    
    Args
    ----------

    tokenizer (transformers.PreTrainedTokenizer):
        The tokenizer used for encoding the data.

    config (transformers.PreTrainedCOnfig):
        The configuration of the model.
    """
    
    name: str = 'MultiLabelClassificationProcessor'
    tokenizer: PreTrainedTokenizer = None
    config: PretrainedConfig = None
    max_length: int = None
    
    def __call__(self, 
                 examples: List[Example],
                ) -> Features:
        
        texts = [example.text for example in examples]
        
        inputs = self.tokenizer(
            texts,
            add_special_tokens=True,
            padding=True,
            truncation=True,
            max_length=self.max_length,
            is_split_into_words=False,
            return_attention_mask=True,
            return_token_type_ids=True,
            return_tensors='tf',
        )
        
        labels = [example.label for example in examples]
        # onehot encoding of empty list yield error
        labels = [[-1] if not l else l for l in labels]
        
        onehot_labels = tf.one_hot(
            tf.ragged.constant(labels),
            depth=self.config.num_labels,
        )
        
        inputs['label'] = tf.reduce_max(onehot_labels, axis=1)
        
        input_features = Features(
            input_ids=inputs['input_ids'],
            attention_mask=inputs['attention_mask'],
            token_type_ids=inputs['token_type_ids'],
            label=inputs['label'],
        )
        
        return input_features

@dataclass
class ClassLabelDataProcessor:
    """
    Data Processor for class-label sequence classification.
    Inputs are dynamically padded to the maximum 
    length of a batch if they are not all of the 
    same length.
    
    Args
    ----------

    tokenizer (transformers.PreTrainedTokenizer):
        The tokenizer used for encoding the data.

    config (transformers.PreTrainedCOnfig):
        The configuration of the model.
    """
    
    name: str = 'ClassLabelDataProcessor'
    tokenizer: PreTrainedTokenizer = None
    config: PretrainedConfig = None
    max_length: int = None
    
    def __call__(self, examples: List[Example]) -> Features:
        
        texts = [(example.text, example.aspect)
                 for example in examples]
        
        inputs = self.tokenizer(
            texts,
            add_special_tokens=True,
            padding=True,
            truncation=True,
            max_length=self.max_length,
            is_split_into_words=False,
            return_attention_mask=True,
            return_token_type_ids=True,
            return_tensors='tf',
        )
        
        inputs['label'] = tf.one_hot(
            [example.label for example in examples],
            depth=self.config.num_labels
        )
        
        input_features = Features(
            input_ids=inputs['input_ids'],
            attention_mask=inputs['attention_mask'],
            token_type_ids=inputs['token_type_ids'],
            label=inputs['label'],
        )
        
        return input_features
    


from transformers import PretrainedConfig
from transformers import TFBertPreTrainedModel 

from nlpipes.layers.bert_layers import TFAdaptedBertMainLayer


class TFAdaptedBertModel(TFBertPreTrainedModel):
    """ The Bert Model with additionnal bottleneck adapters layers 
    outputting raw hidden-states without any specific head on top. """
    def __init__(self, config: PretrainedConfig, *inputs, **kwargs):
        super().__init__(config, *inputs, **kwargs)

        self.bert = TFAdaptedBertMainLayer(config, name="adapted_bert")
    
    def call(self, inputs, **kwargs):
        outputs = self.bert(inputs, **kwargs)

        return outputs



import json
import os
import io
import ijson
import logging
import pprint
from copy import deepcopy
from collections import defaultdict

from dataclasses import dataclass
from typing import Any, List

logger = logging.getLogger(__name__)


@dataclass
class DataConverter():
    
    def from_labelstudio(self, json_file):
        """ """
        task_iterator = self.iterate_on_tasks(json_file)
        for idx, task in enumerate(task_iterator):
            pprint.pprint(task)
            print("\n")
            
    def iterate_on_tasks(self, json_file):
            """ Iterate over the json file of tasks and extract the 
            annotation results"""
            with io.open(json_file, 'rb') as f:
                logger.debug(f'ijson backend in use: {ijson.backend}')
                tasks = ijson.items(
                    f, 'item', use_float=True
                )
                for task in tasks:
                    for annotation in self.get_annotations(task):
                        if annotation is not None:
                            yield annotation
                            
    def get_annotations(self, task):
        """ """
        has_annotations = 'annotations' in task
        
        if not has_annotations:
            logger.warning('Task do not contain annotations')
            return None
        
        annotations = task['annotations']
        
        '''
        # return task with empty annotations
        if not annotations:
            data = Converter.get_data(task, {}, {})
            yield data
        
        # skip cancelled annotations
        cancelled = lambda x: not (
            x.get('skipped', False) or x.get('was_cancelled', False)
        )
        annotations = list(filter(cancelled, annotations))
        if not annotations:
            return None

        # sort by creation time
        annotations = sorted(
            annotations, key=lambda x: x.get('created_at', 0), reverse=True
        )
        '''
        
        data = dict()
        data['text'] = None
        data['labels'] = list()
        labels = dict()
        labels['category'] = None
        labels['polarity'] = None
        labels['aspect'] = None
        labels['opinion'] = None
        
        
        for annotation in annotations:
            result = annotation['result']
            
            for r in result:
                
                if 'from_name' in r and r['from_name'] == 'Categories':
                    value = deepcopy(r['value'])
                    labels['category'] = value['taxonomy']
                    data['text'] = value['text']
                    data['labels'].append(labels)
                    yield data
                    

            #data = Converter.get_data(task, outputs, annotation)
            #data = outputs
            
            #if 'agreement' in task:
            #    data['agreement'] = task['agreement']
            
            


from typing import Callable
from typing import Iterator
from typing import List

from dataclasses import dataclass

import math
import numpy as np

import tensorflow as tf
import transformers

from nlpipes.data.data_types import Example
from nlpipes.data.data_types import Features


""" The Data Loader to load the data in batch.

A obj:`DataLoader` represents an iterable over data 
samples. It represent a batch generator to get a stream
of data reading from a set of examples. Beyond batches
generation, it also apply all the data transformation
to examples into enncoded features that the model can 
understand.
"""

@dataclass
class DataLoader:
    """ An iterable style Data loader wher data loading 
    order is entirely controlled by the user-defined 
    iterable.
    
    Args
    ----------
    name: 
    examples:
    batch_size:
    data_processor:
    max_length:

    """
    
    name: str
    examples: List[Example] = None
    batch_size: int = None
    data_processor: Callable[[List[Example]], Features] = None
    random_seed: int = 69
    
    def __iter__(self) -> Iterator[Features]:
        """ Iterate over the data, shuffles examples 
        and yield a batches of encoded data at 
        each iteration.	""" 
        batch_examples = list()
        indices = np.random.RandomState(self.random_seed).permutation(
            len(self.examples)
        )
        
        for index in indices:
            example = self.examples[index]
            batch_examples.append(example)
            
            if len(batch_examples) == self.batch_size:
                batch_features = self.data_processor(batch_examples)
                batch_examples.clear()
                yield batch_features
    
    def num_batches(self) -> int:
        """ Number of batches per epoch. """
        return math.ceil(len(self.examples) / self.batch_size)
    
    def num_examples(self) -> int:
        """ Number of dataset examples. """
        return len(self.examples)
    


import os
import logging
import shutil

from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple

from dataclasses import dataclass

from pathlib import Path

import tensorflow as tf
import transformers

from tensorflow.keras.optimizers.experimental import AdamW
from tensorflow.keras.metrics import (
    Mean,
    Precision,
    Recall,
    CategoricalAccuracy,
)

from transformers import (
     PretrainedConfig,
     PreTrainedTokenizerFast,
     TFPreTrainedModel,
     TFAutoModelForMaskedLM,
     AutoConfig,
     AutoTokenizer,
)

from tokenizers.pre_tokenizers import Whitespace

from nlpipes.trainers.trainers import Trainer
from nlpipes.losses.losses import SCELoss
from nlpipes.metrics.confusion import F1_Score

from nlpipes.data.data_loaders import DataLoader
from nlpipes.data.data_selectors import DataSelector
from nlpipes.data.data_augmentors import VocabAugmentor
from nlpipes.data.data_processors import SingleLabelProcessor
from nlpipes.data.data_cleaners import clean

from nlpipes.data.data_utils import (
     create_monolabel_examples,
     split_examples,
     generate_batches,
     chunk,
)

from nlpipes.data.data_types import (
     Example,
     Features,
     Prediction,
     Classification,
)

from nlpipes.callbacks.callbacks import (
     TrainingStep,
     History,
     LossHistory,
     AccuracyHistory,
     F1ScoreHistory,
     PrecisionHistory,
     RecallHistory,
     ModelCheckpoint,
     CSVLogger,
     ProgbarLogger,
     EarlyStopping,
     TimedStopping,
)

transformers.logging.set_verbosity_error()

logger = logging.getLogger('__name__')


@dataclass
class PipelineForMonoLabelSequenceClassification():
    """ The pipeline for simple sentiment polarity 
    classification. 
    
    Args
    ----------
    model(TFPreTrainedModel):
       The model used for the task`.
    
    tokenizer(PreTrainedTokenizerFast):
       The rust tokenizer used to convert the input
       text into encoded vectors.
    
    config(PretrainedConfig): 
       The model configuration file.
    """
    
    model: TFPreTrainedModel
    tokenizer: PreTrainedTokenizerFast
    config: PretrainedConfig
    
    def train(
        self, 
        X: List[str], 
        Y: List[str],
        train_frac: float = 0.75,
        test_frac: float = 0.25,
        epochs: int = 4,
        batch_size: int = 16,
        max_length: int = 512,
        learning_rate: float = 2e-5,
        beta_1: float = 0.9,
        beta_2: float = 0.999,
        patience: int = 2,
        min_delta: float = 0.01,
        time_limit: int = 86400,
        random_seed: int = 69,
        verbose: bool = True,
        keep_checkpoints: bool = False,
        checkpoints_dir: str = './model/checkpoints',
        logs_dir: str = './model',
    ) -> Tuple[History]:
        """ Train a model for single-label classification.
        Most of the training process is implemented through 
        callbacks functions. Even the actual training step 
        (calculating the gradient and updating the weights 
        using an optimizer) is implemented as callbacks and
        is not part of the Trainer itself. This provide a 
        modular architecture that allows new ideas to be 
        implemented without having to change and increase
        the complexity of the trainer. """
        
        examples = create_monolabel_examples(
            texts=X, 
            labels=Y,
            config=self.config,
        )
        
        train_examples, test_examples = split_examples(
            examples=examples,
            train_frac=train_frac,
            test_frac=test_frac,
            shuffle=True,
            random_seed=random_seed,
        )        
        
        data_processor = SingleLabelProcessor(
            name='DataProcessor',
            tokenizer=self.tokenizer,
            config=self.config,
            max_length=max_length,
        )
        
        train_stream = DataLoader(
            name='TrainDataLoader',
            examples=train_examples,
            batch_size=batch_size,
            data_processor=data_processor,
            random_seed=random_seed,
        )
        
        test_stream = DataLoader(
            name='TestDataLoader',
            examples=test_examples,
            batch_size=batch_size,
            data_processor=data_processor,
            random_seed=random_seed,
        )
        
        training_step = TrainingStep(
            name='TrainingStep',
            model=self.model,
            loss_fn=SCELoss(),
            optimizer=AdamW(
                learning_rate=learning_rate,
                beta_1=beta_1,
                beta_2=beta_2,
            ),
        )
        
        loss_history = LossHistory(
            name='Loss', 
            training_step=training_step,
            metric=Mean(),
            verbose=verbose,
        )
        
        accuracy_history = AccuracyHistory(
            name='Accuracy',
            training_step=training_step,
            metric=CategoricalAccuracy(),
            verbose=verbose,
        )
        
        f1_history = F1ScoreHistory(
            name='F1',
            training_step=training_step,
            metric=F1_Score(),
            mode='multiclass',
            verbose=verbose,
        )
        
        precision_history = PrecisionHistory(
            name='Precision',
            training_step=training_step,
            metric=Precision(),
            mode='multiclass',
            verbose=verbose,
        )
        
        recall_history = RecallHistory(
            name='Recall',
            training_step=training_step,
            metric=Recall(),
            mode='multiclass',
            verbose=verbose,
        )
        
        model_checkpoint = ModelCheckpoint(
            name='ModelCheckpoint',
            history=loss_history,
            model=self.model, 
            direction='minimize',
            checkpoints_dir=checkpoints_dir,
            verbose=verbose,
        )
        
        early_stopping = EarlyStopping(
            name='EarlyStopping',
            history=loss_history,
            min_delta=min_delta,
            direction='minimize',
            patience=patience,
        )
        
        timed_stopping = TimedStopping(
            name='TimedStopping',
            time_limit=time_limit,
        )
        
        csv_logger = CSVLogger(
            name='CSVLogger',
            logs_dir=logs_dir,
        )
        
        progbar_logger = ProgbarLogger(
            name='ProgbarLogger',
            history=[
                loss_history,
                accuracy_history,
                f1_history,
                precision_history,
                recall_history,
            ],
            batch_size=batch_size,
            num_samples=sum([
                train_stream.num_examples(),
                test_stream.num_examples(),
            ])
        )
        
        callbacks = [
            training_step,
            loss_history,
            accuracy_history,
            f1_history,
            precision_history,
            recall_history,
            model_checkpoint,
            csv_logger,
            progbar_logger,
            early_stopping,
            timed_stopping,
        ]
        
        trainer = Trainer(callbacks)
        
        trainer.fit(
            train_stream=train_stream,
            test_stream=test_stream,
            epochs=epochs,
        )
        
        if keep_checkpoints == False:
            shutil.rmtree(checkpoints_dir)
            
        return (
            loss_history, 
            accuracy_history, 
            f1_history, 
            precision_history, 
            recall_history,
        )
    
    def evaluate(
        self,
        X: List[str],
        Y: List[str],
        metric: tf.metrics.Metric,
        batch_size: int = 32,
    ) -> tf.metrics.Metric:
        """ Evaluate the model on new labeled data """
        
        for texts, labels in generate_batches(X, Y, batch_size):
            examples = self.ingest(texts, labels)
            features = self.transform(examples)
            predictions = self.predict(features)
            
            y_true = [e.label for e in examples]
            y_pred = [p.label for p in predictions]
            
            metric.update_state(y_true, y_pred)
        
        return metric.result()
    
    def classify(
        self, 
        texts: List[str], 
        chunk_text: bool = False,
    ) -> List[Classification]:
        """ Get predictions on new unlabeled data """
         
        examples = self.ingest(texts)
        features = self.transform(examples)
        predictions = self.predict(features)
        classifications = self.postprocess(predictions, examples)
        
        return classifications
    
    def ingest(
        self,
        texts: List[str],
        labels: Optional[str] = None,
        chunk_text: Optional[bool] = False,
    ) -> List[Example]:
        """ Convert each raw input text into examples. 
        Raw texts can optionnaly be chunk into spans. """
        
        if labels:
            label2id = self.config.label2id
            labels = [label2id[label] for label in labels]
            examples = [
                Example(text, label) 
                for text, label 
                in zip(texts, labels)
            ]
        else:
            texts = chunk(text) if chunk_text else texts
            examples = [Example(text) for text in texts] 
        
        return examples
    
    def transform(self, examples: Iterable[Example]) -> Features: 
        """ Transform human-readable examples into encoded
        features that will be used input by the model. """
        
        texts = [e.text for e in examples]
        max_length = self.config.max_position_embeddings

        inputs = self.tokenizer(
           texts,
           add_special_tokens=True,
           padding=True,
           truncation=True,
           max_length=max_length,
           is_split_into_words = False,
           return_attention_mask=True,
           return_token_type_ids=True,
           return_tensors='tf',
        )

        features = Features(
           input_ids=inputs['input_ids'],
           attention_mask=inputs['attention_mask'],
           token_type_ids=inputs['token_type_ids'],
        )

        return features

    def predict(self, features: Features) -> List[Prediction]:
        """ Predict the label index matching the maximum 
        prediction score and compute the loss value
        associated with each prediction. """
        
        outputs = self.model.call(
            input_ids=features.input_ids,
            attention_mask=features.attention_mask,
            token_type_ids=features.token_type_ids,
        )
        
        logits = outputs.logits
        labels = tf.argmax(logits, axis=-1)
        confidences = tf.nn.softmax(logits, axis=1)
        
        predictions = [
            Prediction(label, confidence)
            for label, confidence
            in zip(labels, confidences)
        ]
        
        return predictions

    def postprocess(
        self,
        predictions: Iterable[Prediction],
        examples: Iterable[Example], 
    ) -> List[Classification]:
        """ Post-process the predictions. We convert tensors 
        into more handy python lists and also convert the 
        predicted labels ids into their original name. """
        
        texts = [e.text for e in examples]
        labels = [p.label.numpy() for p in predictions]
        confidences = [p.confidence.numpy() for p in predictions]
        
        labels = [
            self.config.id2label[label] 
            for label in labels
        ]
        
        confidences = [
            max(confidence)
            for confidence in confidences
        ]
        
        classifications = [
            Classification(text, label, confidence)
            for text, label, confidence
            in zip(texts, labels, confidences)
        ]
        
        return classifications
    
    def save(self, save_dir: str):
        """ Save the trained model on disk """
        
        os.makedirs(save_dir, exist_ok=True)
        self.model.save_pretrained(save_dir)
        self.tokenizer.save_pretrained(save_dir)
        self.config.save_pretrained(save_dir)



import os
import logging
import shutil

from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple

from dataclasses import dataclass

from pathlib import Path

import tensorflow as tf
import transformers

from tensorflow.keras.optimizers.experimental import AdamW
from tensorflow.keras.metrics import (
    Mean,
    Precision,
    Recall,
    CategoricalAccuracy,
)

from transformers import (
    PretrainedConfig,
    PreTrainedTokenizerFast,
    TFPreTrainedModel,
    TFAutoModelForMaskedLM,
    AutoConfig,
    AutoTokenizer,
)

from tokenizers.pre_tokenizers import Whitespace

from nlpipes.trainers.trainers import Trainer
from nlpipes.losses.losses import SCELoss
from nlpipes.metrics.confusion import F1_Score

from nlpipes.data.data_loaders import DataLoader
from nlpipes.data.data_selectors import DataSelector
from nlpipes.data.data_augmentors import VocabAugmentor
from nlpipes.data.data_processors import ClassLabelDataProcessor
from nlpipes.data.data_cleaners import clean

from nlpipes.data.data_utils import (
     create_classlabel_examples,
     generate_batches,
     get_labels_depth,
     split_examples,
     make_alignment,
     chunk,
)

from nlpipes.data.data_types import (
     Example,
     Features,
     Prediction,
     Classification,
)

from nlpipes.callbacks.callbacks import (
     TrainingStep,
     History,
     LossHistory,
     AccuracyHistory,
     F1ScoreHistory,
     PrecisionHistory,
     RecallHistory,
     ModelCheckpoint,
     CSVLogger,
     ProgbarLogger,
     EarlyStopping,
     TimedStopping,
)

transformers.logging.set_verbosity_error()

logger = logging.getLogger('__name__')


@dataclass
class PipelineForClassLabelSequenceClassification():
    """ The pipeline for aspect sentiment classification.
    
    Args
    ----------
    model(TFPreTrainedModel):
       The model used for the task`.
    
    tokenizer(PreTrainedTokenizerFast):
       The rust tokenizer used to convert the input
       text into encoded vectors.
    
    config(PretrainedConfig): 
       The model configuration file.
    """
    
    model: TFPreTrainedModel
    tokenizer: PreTrainedTokenizerFast
    config: PretrainedConfig
    
    def train(self, 
              X: List[str], 
              Y: List[str],
              train_frac: float = 0.80,
              test_frac: float = 0.20,
              epochs: int = 3,
              batch_size: int = 16,
              max_length: int = 128,
              lr: float = 2e-5,
              beta_1: float = 0.9,
              beta_2: float = 0.999,
              patience: int = 2,
              min_delta: float = 0.01,
              time_limit: int = 86400,
              random_seed: int = 69,
              verbose: bool = True,
              keep_checkpoints: bool = False,
              checkpoints_dir: str = './model/checkpoints',
              logs_dir: str = './model',
             ) -> Tuple[History]:
        """ Train a model for class-label sequence classification """
        
        examples = create_classlabel_examples(
            texts=X, 
            labels=Y,
            config=self.config,
        )
        
        train_examples, test_examples = split_examples(
            examples,
            train_frac=train_frac,
            test_frac=test_frac,
            shuffle=True,
            random_seed=random_seed,
        )
        
        data_processor = ClassLabelDataProcessor(
            name='DataProcessor',
            tokenizer=self.tokenizer,
            config=self.config,
            max_length=max_length,
        )
        
        train_stream = DataLoader(
            name='TrainDataLoader',
            examples=train_examples,
            batch_size=batch_size,
            data_processor=data_processor,
            random_seed=random_seed,
        )
        
        test_stream = DataLoader(
            name='TestDataLoader',
            examples=test_examples,
            batch_size=batch_size,
            data_processor=data_processor,
            random_seed=random_seed,
        )
        
        training_step = TrainingStep(
            name='TrainingStep',
            model=self.model,
            loss_fn=SCELoss(),
            optimizer=AdamW(
                learning_rate=lr,
                beta_1=beta_1,
                beta_2=beta_2,
            ),
        )
        
        loss_history = LossHistory(
            name='Loss', 
            metric=Mean(),
            training_step=training_step,
            verbose=verbose,
        )
        
        accuracy_history = AccuracyHistory(
            name='Accuracy',
            metric=CategoricalAccuracy(),
            training_step=training_step,
            verbose=verbose,
        )
        
        f1_history = F1ScoreHistory(
            name='F1',
            training_step=training_step,
            metric=F1_Score(),
            mode='multiclass',
            verbose=verbose,
        )
        
        precision_history = PrecisionHistory(
            name='Precision',
            training_step=training_step,
            metric=Precision(),
            mode='multiclass',
            verbose=verbose,
        )
        
        recall_history = RecallHistory(
            name='Recall',
            training_step=training_step,
            metric=Recall(),
            mode='multiclass',
            verbose=verbose,
        )
        
        model_checkpoint = ModelCheckpoint(
            name='ModelCheckpoint',
            history=loss_history,
            model=self.model, 
            direction='minimize',
            checkpoints_dir=checkpoints_dir,
            verbose=verbose,
        )
        
        early_stopping = EarlyStopping(
            name='EarlyStopping',
            history=loss_history,
            min_delta=min_delta,
            direction='minimize',
            patience=patience,
        )
        
        timed_stopping = TimedStopping(
            name='TimedStopping',
            time_limit=time_limit,
        )
        
        csv_logger = CSVLogger(
            name='CSVLogger',
            logs_dir=logs_dir,
        )
        
        progbar_logger = ProgbarLogger(
            name='ProgbarLogger',
            batch_size=batch_size,
            history=[
                loss_history,
                accuracy_history,
                f1_history,
                precision_history,
                recall_history,
            ],                
            num_samples=sum([
                train_stream.num_examples(),
                test_stream.num_examples(),
            ]))
        
        callbacks = [
            training_step,
            loss_history,
            accuracy_history,
            f1_history,
            precision_history,
            recall_history,
            model_checkpoint,
            csv_logger,
            progbar_logger,
        ]
        
        trainer = Trainer(callbacks)
        
        trainer.fit(
            train_stream=train_stream,
            test_stream=test_stream,
            epochs=epochs,
        )
        
        if keep_checkpoints == False:
            shutil.rmtree(checkpoints_dir)
            
        return (
            loss_history, 
            accuracy_history, 
            fbeta_history, 
            precision_history, 
            recall_history,
        )
    
    def evaluate(self,
                 X: List[str],
                 Y: List[str],
                 metric: tf.metrics.Metric,
                 batch_size: int = 16,
                ) -> tf.metrics.Metric:
        """ Evaluate the model performance on new data
        according to the user-defined evaluation metric """
        
        for texts, labels in generate_batches(X, Y, batch_size):
            examples = self.ingest(texts, labels=labels)
            features = self.transform(examples)
            predictions = self.predict(features)
            
            y_true = [e.label for e in examples]
            y_pred = [p.label for p in predictions]
            
            metric.update_state(y_true, y_pred)
        
        return metric.result()
    
    def classify(
        self, 
        texts: List[str],
        aspects: List[List[str]],
    ) -> List[Classification]:
        """ Get classification on new unlabeled data """
         
        examples = self.ingest(texts, aspects)
        features = self.transform(examples)
        predictions = self.predict(features)
        classifications = self.postprocess(
            predictions, examples, texts
        )
        
        return classifications
        
    def ingest(
        self,
        texts: List[str],
        aspects: Optional[List[str]] = None,
        labels: Optional[List[str]] = None,
        chunk_text: Optional[bool] = True,
    ) -> List[Example]:
        """ Serialize each raw text into examples. 
        We keep the a lignment between the examples and 
        the original raw input texts. """
        
        examples = []
        
        if labels:
            for idx, text in enumerate(texts):
                aspects = [label[0] for label in labels[idx]]
                aspect_labels = [label[1] for label in labels[idx]]

                for num, aspect in enumerate(aspects):
                    aspect = aspect.lower()
                    label = aspect_labels[num]
                    label = self.config.label2id[label]
                    example = Example(
                        text=text,
                        aspect=aspect,
                        label=label,
                        alignment=idx,
                    )
                    examples.append(example)
                    
        else:
            for idx, text in enumerate(texts):
                for aspect in aspects[idx]:
                    example = Example(
                        text=text, 
                        aspect=aspect, 
                        alignment=idx,
                    )
                    examples.append(example)
                    
        return examples
    
    def transform(self, examples: Iterable[Example]) -> Features: 
        """ Transform human-readable sequences into encoded
        features that can be use as input by the model. """
        
        text_pairs = [(e.text, e.aspect) for e in examples]
        max_length = self.config.max_position_embeddings
        
        inputs = self.tokenizer(
           text_pairs,
           add_special_tokens=True,
           padding=True,
           truncation=True,
           max_length=max_length,
           is_split_into_words = False,
           return_attention_mask=True,
           return_token_type_ids=True,
           return_tensors='tf',
        )

        features = Features(
           input_ids=inputs['input_ids'],
           attention_mask=inputs['attention_mask'],
           token_type_ids=inputs['token_type_ids'],
        )
        
        return features
    
    def predict(self, features: Features) -> List[Prediction]:
        """ Predict the label index matching the maximum
        prediction score and compute the loss value associated
        with each prediction. """

        outputs = self.model.call(
            input_ids=features.input_ids,
            attention_mask=features.attention_mask,
            token_type_ids=features.token_type_ids,
        )

        logits = outputs.logits
        confidences = tf.nn.softmax(logits, axis=1)
        labels = tf.argmax(logits, axis=-1)
        
        predictions = [
            Prediction(label, confidence)
            for label, confidence
            in zip(labels, confidences)
        ]
        
        return predictions
    
    def postprocess(
        self, 
        predictions: Iterable[Prediction],
        examples: Iterable[Example],
        texts: Iterable[str], 
    ) -> List[Classification]:
        """ Post-process the predictions. We convert tensors 
        into more convenient python lists and also convert the 
        predicted labels ids into their original name. """
        
        aspects = [e.aspect for e in examples]
        alignments = [e.alignment for e in examples]
        
        labels = [p.label.numpy() for p in predictions]
        labels = [self.config.id2label[l] for l in labels]
        labels = [[a, l] for a, l in zip(aspects, labels)]
        labels = make_alignment(labels, alignments)
        
        confidences = [p.confidence.numpy() for p in predictions]
        confidences = [max(c.tolist()) for c in confidences]
        confidences = make_alignment(confidences, alignments)
        
        classifications = [
            Classification(text, label, confidence)
            for text, label, confidence
            in zip(texts, labels, confidences)
        ]

        return classifications
        
    def save(self, save_dir: str):
        """ Save the trained model on disk """
        
        os.makedirs(save_dir, exist_ok=True)
        self.model.save_pretrained(save_dir)
        self.tokenizer.save_pretrained(save_dir)
        self.config.save_pretrained(save_dir)
        


"""Diversity metrics for for in-domain subcorpus selection.
Code adapted from: 
github: https://github.com/georgian-io/Transformers-Domain-Adaptation.
paper: https://github.com/sebastianruder/learn-to-select-data.
"""

from functools import partial
from typing import Callable, Dict, Sequence, NewType
from typing_extensions import Literal

import numpy as np
import scipy.stats

from nlpipes.data.data_types import Token


def entropy(example: Sequence[Token], 
            vocab2id: Dict[Token, int],
           ) -> float:
    """Calculate Entropy 
    ref: https://en.wikipedia.org/wiki/Entropy_(information_theory%29#Definition).
    """
    example = {term for term in example if term in vocab2id}
    term_ids = [vocab2id[term] for term in example]
    return scipy.stats.entropy(term_ids)

def simpsons_index(
    example: Sequence[Token], 
    train_term_dist: np.ndarray, 
    vocab2id: Dict[Token, int],
) -> float:
    """ Calculate Simpson's Index 
    ref: https://en.wikipedia.org/wiki/Diversity_index#Simpson_index
    """
    if not len(example):
        return 0
    example = {term for term in example if term in vocab2id}
    term_ids = [vocab2id[term] for term in example]
    score = (train_term_dist[term_ids] ** 2).sum()
    return score

def renyi_entropy(
    example: Sequence[Token], 
    domain_tokens_distribution: np.ndarray, 
    vocab2id: Dict[Token, int],
) -> float:
    """ Calculate Rényi Entropy
    ref: https://en.wikipedia.org/wiki/R%C3%A9nyi_entropy
    """
    example = {term for term in example if term in vocab2id}
    term_ids = [vocab2id[term] for term in example]

    alpha = 0.99
    summed = (domain_tokens_distribution[term_ids] ** alpha).sum()
    if summed == 0:
        # 0 if none of the terms appear in the dictionary;
        # set to a small constant == low prob instead
        summed = 0.0001
    score = 1 / (1 - alpha) * np.log(summed)
    return score

def number_of_term_types(example: Sequence[Token]) -> int:
    """Calculate the number of term types of the example."""
    return len(set(example))

def type_token_diversity(example: Sequence[Token]) -> float:
    """Calculate diversity based on the type-token ratio of the example."""
    if not len(example):
        return 1
    type_token_ratio = number_of_term_types(example) / len(example)
    return -type_token_ratio


DiversityMetric = Literal[
    "num_token_types",
    "type_token_ratio",
    "entropy",
    "simpsons_index",
    "renyi_entropy",
]
DiversityFunction = Callable[
    [Sequence[Token]], float
]
DIVERSITY_FEATURES = {
    "num_token_types",
    "type_token_ratio",
    "entropy",
    "simpsons_index",
    "renyi_entropy",
}


def diversity_metrics_factory(
    metric: DiversityMetric, 
    tokens_distribution: np.ndarray, 
    vocab2id: Dict[Token, int]
) -> DiversityFunction:
    """Return the corresponding diversity function based on the provided metric.
    Args:
        metric (str): Diversity metric
        train_term_dist: Term distribution of the training data
        vocab2id: Vocabulary-to-id mapping
    Raises:
        ValueError: If `metric` does not exist in DIVERSITY_FEATURES
    """
    if metric not in DIVERSITY_FEATURES:
        raise ValueError(f'"{metric}" is not a valid diversity metric.')

    mapping: Dict[DiversityMetric, DiversityFunction] = {
        "num_token_types": number_of_term_types,
        "type_token_ratio": type_token_diversity,
        "entropy": partial(entropy, vocab2id=vocab2id),
        "simpsons_index": partial(
            simpsons_index, 
            tokens_distribution=tokens_distribution,
            vocab2id=vocab2id,
        ),
        "renyi_entropy": partial(
            renyi_entropy,
            domain_tokens_distribution=tokens_distribution,
            vocab2id=vocab2id,
        ),
    }
    
    diversity_function = mapping[metric]
    return diversity_function



import tensorflow as tf


class LearningRateSchedulerWithWarmup(tf.optimizers.schedules.LearningRateSchedule):
    
    """ A custom learning rate scheduler for optimizer according to the
    formula in the paper `Attention Is All You Need`.
    Reference: 
    - https://arxiv.org/abs/1706.03762
    """
    
    def __init__(self, d_model=128, warmup_steps=4000):
        super(LearningRateWithWarmup, self).__init__()
        self.d_model = d_model
        self.d_model = tf.cast(self.d_model, tf.float32)
        self.warmup_steps = warmup_steps
    
    def __call__(self, step):
        arg1 = tf.math.rsqrt(step)
        arg2 = step * (self.warmup_steps ** -1.5)
        learning_rate = tf.math.rsqrt(self.d_model) * tf.math.minimum(arg1, arg2)
        return learning_rate 



import logging
from enum import IntEnum

from collections import OrderedDict

from dataclasses import dataclass
from dataclasses import fields
from dataclasses_json import dataclass_json

from typing import Any
from typing import Dict
from typing import Iterable
from typing import List
from typing import NewType
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import Union

import numpy as np
import tensorflow as tf

from transformers import (
    PreTrainedModel,
    PreTrainedTokenizer,
    PretrainedConfig,
)

from nlpipes.pipelines.base_pipeline import Pipeline


Document = NewType("Document", str)
Corpus = NewType("Corpus", Sequence[Document])
Token = NewType("Token", str)
Sequence = NewType("Sequence", str)
AllLabel = NewType("Label", str)


@dataclass
class Task:
    """ """
    pipeline: Pipeline
    model: PreTrainedModel
    tokenizer: PreTrainedTokenizer
    config: PretrainedConfig 
    
    
@dataclass
class Example:
    """ A single example used to create a classifier.
        
    Args
    ----------
        text: raw input text.
        aspects: (optional): to be specified when the labels is 
                 toward a specific aspect in the input text.
        label: (optional) target label.
    """
    text: str
    label: Optional[str] = None
    aspect: Optional[str] = None
    alignment: Optional[int] = None  
    
@dataclass
class Tokens:
    """ A set of tokens to be encoded to be used
    as input by the transformer model.
    
    Args
    ----------
        words: words split into words using a 
            `whitespace` tokenizer.
        subwords: words split into subwords
             using a model specific tokenizer.
        subwords_plus: subwords with the additionnal 
             special tokens '[CLS]' and '[SEP]'.
    """
    words: List[str]
    subwords: List[str]
    subwords_plus: List[str]
    
    
@dataclass
class Features:
    """ A set of features expected as input by a transformer model.
    
    Args
    ----------
        input_ids: indices of input sequence tokens in the vocabulary.
        attention_mask: mask to avoid performing attention on padding 
                        token indices. Mask values selected in `[0, 1]`.
                        Usually `1` for tokens that are NOT MASKED and 
                        `0` for MASKED (padded) tokens.
        token_type_ids: segment token indices to indicate first and 
                        second portions of the inputs. Only some models 
                        use them.
        label: (Optional) the target label corresponding to the input.
    """
    input_ids: tf.Tensor
    attention_mask: tf.Tensor = None
    token_type_ids: tf.Tensor = None
    label: Optional[tf.Tensor] = None
    
    
@dataclass
class Prediction:
    """ The model predictions 

    Args
    ----------
        label:
        confidence:
    """
    label: tf.Tensor
    confidence: tf.Tensor
    
@dataclass_json  
@dataclass
class Classification:
    """ The final outcome after post-processing the 
    model outputs

    Args
    ----------
        text:
        prediction:
        confidence:
    """
    text: List[str]
    prediction: List[str]
    confidence: List[str]
    
    
@dataclass
class TFSequenceClassifierOutput():
    """ The model outputs for a single example.
    
    Args
    ----------
        logits: the logits scores outputed by the
                classification head.
        hidden_states: hidden-states of the model at the output of each
                layer plus the initial embedding outputs.
        attentions: attentions weights after the attention softmax, used
                to compute the weighted average in the self-attention 
                heads.
    """
    logits: tf.Tensor
    hidden_states: Optional[Tuple[tf.Tensor]]
    attentions: Optional[Tuple[tf.Tensor]]


@dataclass
class TFSequenceClassifierBatchOutput(TFSequenceClassifierOutput):
    loss: tf.Tensor
    
    
@dataclass
class LabelsMapping:
    """ The labels used for classification tasks. """
    def __init__(self, labels):
        for idx, label in enumerate(labels):
            setattr(self, label, idx)
    def id2label(self):
        return {idx:label 
                for label, idx in self.__dict__.items()}
    def label2id(self):
        return {label:idx 
                for label, idx  in self.__dict__.items()}
    def num_labels(self):
        return len(self.__dict__)
    
    
# Source: https://github.com/huggingface/transformers
# TODO: clean unrelevant code
@dataclass
class ModelOutput(OrderedDict):
    """
    Base class for all model outputs as dataclass. Has a `__getitem__` 
    that allows indexing by integer or slice (like a tuple) or strings
    (like a dictionary) that will ignore the `None` attributes. Otherwise
    behaves like a regular python dictionary.
    <Tip warning={true}>
    You can't unpack a `ModelOutput` directly. Use the 
    [`~utils.ModelOutput.to_tuple`]	method to convert it to a tuple before.
    </Tip>
    """

    def __post_init__(self):
        class_fields = fields(self)

        # Safety and consistency checks
        if not len(class_fields):
            raise ValueError(f"{self.__class__.__name__} has no fields.")
        if not all(field.default is None for field in class_fields[1:]):
            raise ValueError(f"{self.__class__.__name__} should not have more than one required field.")

        first_field = getattr(self, class_fields[0].name)
        other_fields_are_none = all(getattr(self, field.name) is None for field in class_fields[1:])

        if other_fields_are_none and not isinstance(first_field, tf.Tensor):
            if isinstance(first_field, dict):
                iterator = first_field.items()
                first_field_iterator = True
            else:
                try:
                    iterator = iter(first_field)
                    first_field_iterator = True
                except TypeError:
                    first_field_iterator = False

            if first_field_iterator:
                for element in iterator:
                    if (
                        not isinstance(element, (list, tuple))
                        or not len(element) == 2
                        or not isinstance(element[0], str)
                    ):
                        break
                    setattr(self, element[0], element[1])
                    if element[1] is not None:
                        self[element[0]] = element[1]
            elif first_field is not None:
                self[class_fields[0].name] = first_field
        else:
            for field in class_fields:
                v = getattr(self, field.name)
                if v is not None:
                    self[field.name] = v

    def __delitem__(self, *args, **kwargs):
        raise Exception(f"You cannot use ``__delitem__`` on a {self.__class__.__name__} instance.")

    def setdefault(self, *args, **kwargs):
        raise Exception(f"You cannot use ``setdefault`` on a {self.__class__.__name__} instance.")

    def pop(self, *args, **kwargs):
        raise Exception(f"You cannot use ``pop`` on a {self.__class__.__name__} instance.")

    def update(self, *args, **kwargs):
        raise Exception(f"You cannot use ``update`` on a {self.__class__.__name__} instance.")

    def __getitem__(self, k):
        if isinstance(k, str):
            inner_dict = {k: v for (k, v) in self.items()}
            return inner_dict[k]
        else:
            return self.to_tuple()[k]

    def __setattr__(self, name, value):
        if name in self.keys() and value is not None:
            super().__setitem__(name, value)
        super().__setattr__(name, value)

    def __setitem__(self, key, value):
        super().__setitem__(key, value)
        super().__setattr__(key, value)

    def to_tuple(self) -> Tuple[Any]:
        """ Convert self to a tuple containing all the attributes/keys that
        are not `None`. """
        return tuple(self[k] for k in self.keys())


@dataclass
class TFBaseModelOutput(ModelOutput):
    """ Base class for model's outputs, with potential hidden states and
    attentions. """

    last_hidden_state: tf.Tensor = None
    hidden_states: Optional[Tuple[tf.Tensor]] = None
    attentions: Optional[Tuple[tf.Tensor]] = None


@dataclass
class TFBaseModelOutputWithPooling(ModelOutput):
    """ Base class for model's outputs that also contains a pooling of 
    the last hidden states. """

    last_hidden_state: tf.Tensor = None
    pooler_output: tf.Tensor = None
    hidden_states: Optional[Tuple[tf.Tensor]] = None
    attentions: Optional[Tuple[tf.Tensor]] = None



import os
import logging
import shutil

from typing import List
from typing import Optional
from typing import Union

from pathlib import Path

from dataclasses import dataclass

import tensorflow as tf
import transformers
from tensorflow.keras.optimizers.experimental import AdamW

from transformers import (
    PretrainedConfig,
    PreTrainedTokenizer,
    TFPreTrainedModel,
)

from nlpipes.trainers.trainers import Trainer

from nlpipes.losses.losses import SCELoss

from nlpipes.data.data_loaders import DataLoader
from nlpipes.data.data_processors import MLMProcessor

from nlpipes.data.data_selectors import DataSelector
from nlpipes.data.data_augmentors import VocabAugmentor
from nlpipes.data.data_cleaners import clean

from nlpipes.data.data_utils import (
     create_examples,
     split_examples,
)

from nlpipes.data.data_types import (
     Document,
     Corpus,
     Token,
     Example,
)

from nlpipes.callbacks.callbacks import (
     TrainingStep,
     History,
     LossHistory,
     AccuracyHistory,
     ModelCheckpoint,
     CSVLogger,
     ProgbarLogger,
     EarlyStopping,
     TimedStopping,
)

transformers.logging.set_verbosity_error()

logger = logging.getLogger('__name__')


@dataclass
class PipelineForMaskedLanguageModeling():
    """ The `pipeline` for the masked language modeling.
    
    Args
    ----------
    model(TFPreTrainedModel):
       The model used for the task.
    
    tokenizer(PreTrainedTokenizerFast):
       The rust tokenizer used to convert the input
       text into encoded vectors.
    
    config(PretrainedConfig): 
       The model configuration file.
    """
    
    model: TFPreTrainedModel
    tokenizer: PreTrainedTokenizer
    config: PretrainedConfig
    
    def train(
        self,
        texts: Corpus,
        train_frac: float = 0.80,
        test_frac: float = 0.20,
        epochs: int = 4,
        batch_size: int = 8,
        max_length: int = 256,
        learning_rate: float = 2e-5,
        beta_1: float = 0.9,
        beta_2: float = 0.999,
        mlm_prob: float = 0.15,
        target_corpus_size: Union[int, float] = 0.8,
        target_vocab_size: int = 31000,
        similarity_metrics: Optional[List[str]] = ['euclidean'],
        diversity_metrics: Optional[List[str]] = ['entropy'],
        min_delta: float = 0.01,
        patience: int = 5,
        time_limit: int = 86400,
        random_seed: int = 69,
        keep_checkpoints: bool = False,
        verbose: bool = True,
        save_corpus_dir: str = './model',
        checkpoints_dir: str = './model/checkpoints',
        logs_dir: str = './model',
    ) -> History:
        """ Train the model on a masked modeling task """
        
        data_selector = DataSelector(
            name='DataSelector',
            tokenizer=self.tokenizer,
            target_corpus_size=target_corpus_size,
            similarity_metrics=similarity_metrics,
            diversity_metrics=diversity_metrics,
        )
        
        vocab_augmentor = VocabAugmentor(
            name='VocabAugmentor',
            tokenizer=self.tokenizer,
            target_vocab_size=target_vocab_size,
            cased=False,
        )
        
        clean_texts = [clean(text) for text in texts]
        selected_texts = data_selector(clean_texts)
        new_tokens = vocab_augmentor.get_new_tokens(selected_texts)
        self.tokenizer.add_tokens(new_tokens)
        self.model.resize_token_embeddings(len(self.tokenizer))
        
        examples = create_examples(selected_texts)
        
        train_examples, test_examples = split_examples(
            examples=examples,
            train_frac=train_frac,
            test_frac=test_frac, 
            shufffle=True, 
        )

        data_processor = MLMProcessor(
            name='DataProcessor',
            tokenizer=self.tokenizer,
            config=self.config,
            max_length=max_length,
            mlm_prob=mlm_prob,
            mask_prob=0.80,
            rand_prob=0.10,
            idem_prob=0.10,
        )
        
        train_stream = DataLoader(
            name='TrainDataLoader',
            examples=train_examples,
            batch_size=batch_size,
            data_processor=data_processor,
        )
        
        test_stream = DataLoader(
            name='TestDataLoader',
            examples=test_examples,
            batch_size=batch_size,
            data_processor=data_processor,
        )
        
        training_step = TrainingStep(
            name='TrainingStep',
            model=self.model,
            loss_fn=SCELoss(),
            optimizer=AdamW(
                learning_rate=learning_rate,
                beta_1=beta_1,
                beta_2=beta_2,
            ),
        )
        
        loss_history = LossHistory(
            name='Loss', 
            metric = tf.metrics.Mean(),
            training_step=training_step,
            verbose=verbose,
        )
        
        model_checkpoint = ModelCheckpoint(
            name='ModelCheckpoint',
            history=loss_history,
            model=self.model, 
            direction='minimize',
            checkpoints_dir=checkpoints_dir,
            verbose=verbose,
        )
        
        early_stopping = EarlyStopping(
            name='EarlyStopping',
            history=loss_history,
            direction='minimize',
            min_delta=min_delta,
            patience=patience,
        )
        
        timed_stopping = TimedStopping(
            name='TimedStopping',
            time_limit=time_limit,
        )
        
        csv_logger = CSVLogger(
            name='CSVLogger',
            logs_dir=logs_dir,
        )
        
        progbar_logger = ProgbarLogger(
            name='ProgbarLogger',
            batch_size=batch_size,
            history=[
                loss_history,
            ],
            num_samples=sum([
                train_stream.num_examples(),
                test_stream.num_examples(),
            ])
        )
        
        callbacks = [
            training_step,
            loss_history,
            model_checkpoint,
            csv_logger,
            progbar_logger,
            early_stopping,
            timed_stopping,
        ]
        
        trainer = Trainer(callbacks)
        
        trainer.fit(
            train_stream=train_stream,
            test_stream=test_stream,
            epochs=epochs,
        )
        
        if keep_checkpoints == False:
            shutil.rmtree(checkpoints_dir)
            
        os.makedirs(save_corpus_dir, exist_ok=True) 
        corpus_path = os.path.join(save_corpus_dir, 'selected_corpus.txt')         
        Path(corpus_path).write_text('\n'.join(selected_texts))
        
        return loss_history
    
    def evaluate(self):
        """ Evaluate the model performance on new data
        according to the user-defined evaluation metric """
        raise NotImplementedError
    
    def save(self, save_dir: str):
        """ Save the model and the tokenizer on disk """
        
        os.makedirs(save_dir, exist_ok=True)        
        self.model.save_pretrained(save_dir)
        self.tokenizer.save_pretrained(save_dir)
        self.config.save_pretrained(save_dir)
        


"""Text-related util functions."""
from ftfy import fix_text
import re

def clean(text: str, lowercase: bool = False) -> str:
    """Clean text. Does the following:
        1. Replace whitespace characters with space
        2. Fix corrupted text encodings
        3. Lowercase (if specified)
        4. Remove non latin character
    """
    # Replace whitespace characters with spaces
    text = " ".join(text.split())

    # Fix encoded texts
    text = fix_text(text)

    # Lowercase text if specified
    if lowercase:
        text = text.lower()
        
    # Remove non latin character
    text = re.sub(r'[^\x00-\x7f]',r'', text)

    return text


import os
import sys
import copy
import logging
import time
import datetime
import dataclasses

from abc import ABC, abstractmethod
from collections import defaultdict
from tqdm.auto import tqdm

import numpy as np
import tensorflow as tf
import transformers

from typing import (
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
)

from dataclasses import (
    astuple,
    dataclass,
    field,
)

from nlpipes.data.data_types import (
    Features,
    TFSequenceClassifierBatchOutput,
)

logger = logging.getLogger('__name__')


"""
`Callbacks` encapsulate a set of functions to be applied at given
stages of the training procedure to customize the trainer behavior. 
The relevant callbacks functions will then be called at the 
various stages of the model training lifecycle, such as when:
 - the training starts.
 - the training ends.
 - the testing starts.
 - the testing ends.
 - an epoch starts
 - an epoch ends.
 - a batch starts.
 - a batch ends.
"""

class Callback(ABC):
    
    """ Abstract base class for callback objects. """
    
    def on_train_begin(self):
        """ Called at the beginning of the training. """
        
    def on_train_end(self):
        """ Called at the end of the training. """
        
    def on_test_begin(self):
        """ Called at the beginning of the evaluation. """
        
    def on_test_end(self):
        """ Called at the end of the evaluation. """
    
    def on_epoch_begin(self, *args):
        """ Called at the beginning of an epoch. """
        
    def on_epoch_end(self, *args):
        """ Called at the end of an epoch. """
    
    def on_train_batch_begin(self, *args):
        """ Called at the beginning of a training batch. """
    
    def on_train_batch_end(self, *args):
        """ Called at the end of a training batch. """
    
    def on_test_batch_begin(self, *args):
        """ Called at the beginning of a test batch. """

    def on_test_batch_end(self, *args):
        """ Called at the end of a testing batch. """


@dataclass
class CallbackList(Callback):
    """ Internal class that just calls the list of callbacks.
    The CallbackList serves two main purposes:
        1. It stores all the callbacks.
        2. It allows to call all of the individual callbacks 
           easily. For example, if we have three callbacks 
           that do something at the end of an epoch, then 
           callbacks.on_epoch_end() will call the three
           callback objects in order. 
    """

    callbacks: List[Callback]
    
    def on_train_begin(self, *args):
        for callback in self.callbacks:
            callback.on_train_begin(*args)
        
    def on_train_end(self, *args):
        for callback in self.callbacks:
            callback.on_train_end(*args)
        
    def on_test_begin(self, *args):
        for callback in self.callbacks:
            callback.on_test_begin(*args)
        
    def on_test_end(self, *args):
        for callback in self.callbacks:
            callback.on_test_end(*args)

    def on_epoch_begin(self, *args):
        for callback in self.callbacks:
            callback.on_epoch_begin(*args)

    def on_epoch_end(self, *args):
        for callback in self.callbacks:
            callback.on_epoch_end(*args)
            
    def on_train_batch_begin(self, *args):
        for callback in self.callbacks:
            callback.on_train_batch_begin(*args)

    def on_train_batch_end(self, *args):
        for callback in self.callbacks:
            callback.on_train_batch_end(*args)

    def on_test_batch_begin(self, *args):
        for callback in self.callbacks:
            callback.on_test_batch_begin(*args)

    def on_test_batch_end(self, *args):
        for callback in self.callbacks:
            callback.on_test_batch_end(*args)
            
            
@dataclass
class TrainingStep(Callback):
    """ Callback that define the training step. The training step is
    a single iteration of the training process. It consists of a 
    forward pass (where the obj:`model` makes predictions based on
    the input data) and a backward pass (where the obj:`optimizer`
    updates the obj:`model` weights and biases based on the difference
    between the predicted output and the expected output as expressed
    by the obj:`loss_fn`).
    """
    
    name: str = 'TrainingStep'
    model: transformers.TFPreTrainedModel = None
    optimizer: tf.optimizers.Optimizer = None
    loss_fn: tf.losses = None
    
    def train_step(self, input_batch: Features):
        """ Train the model on a single batch of examples. """
        
        with tf.GradientTape() as tape:
            output_batch = self.model.call(
                input_ids=input_batch.input_ids,
                attention_mask=input_batch.attention_mask,
                token_type_ids=input_batch.token_type_ids,
                training=True,
            )
            
            loss = self.loss_fn(
                input_batch.label, output_batch.logits
            )
            
        variables = self.model.trainable_variables
        gradients = tape.gradient(loss, variables)
        
        self.optimizer.apply_gradients(
            zip(gradients, variables)
        )
        
        return TFSequenceClassifierBatchOutput(
            loss=loss,
            logits=output_batch.logits,
            hidden_states=output_batch.hidden_states,
            attentions=output_batch.attentions,
        )
    
    def test_step(self, input_batch: Features):
        """ Test the model on a single batch of examples. """
        
        output_batch = self.model.call(
            input_ids=input_batch.input_ids,
            attention_mask=input_batch.attention_mask,
            token_type_ids=input_batch.token_type_ids,
            training=False,
        )
        
        loss = self.loss_fn(
            input_batch.label, output_batch.logits
        )
        
        return TFSequenceClassifierBatchOutput(
            loss=loss,
            logits=output_batch.logits,
            hidden_states=output_batch.hidden_states,
            attentions=output_batch.attentions,
        )
    
    def on_train_begin(self):
        message = self.model.summary()
        print(message)
        
    def on_train_batch_begin(self, step: int, batch: Features):
        self.train_step_output = self.train_step(batch)
             
    def on_test_batch_begin(self, step: int, batch: Features):
        self.test_step_output = self.test_step(batch)


@dataclass
class History(Callback):
    """ Callback that record the performance of the model 
    during training. It stores all the evaluation metrics 
    such as loss and accuracy at each training and testing
    steps. This allows to track the progress of the model
    over time, and can be used by other callbacks, e.g. to
    detect when the model has reached its optimal performance.
    """
    
    name: str
    metric: Callable[[], tf.metrics.Metric]
    train: Dict = field(default_factory=dict)
    test: Dict = field(default_factory=dict)
    epoch: int = 0
    verbose: bool = False 
    
    def __post_init__(self):
        """ """
        self.train_metric = copy.deepcopy(self.metric)
        self.test_metric = copy.deepcopy(self.metric)
        
    def on_epoch_begin(self, epoch: int):
        """ Resets all of the metric state variables """
        self.epoch = epoch
        self.train_metric.reset_states()
        self.test_metric.reset_states()
        
    def on_epoch_end(self, epoch: int):
        """ Display the evaluation metrics after each epoch """        
        self.train[epoch] = self.train_metric.result()
        self.test[epoch] = self.test_metric.result()
        
        if self.verbose:
            message = f'Epoch {epoch+1:4d} {self.name:10} ' \
                      f'Train {self.train[epoch]:0.6f} ' \
                      f'Test  {self.test[epoch]:0.6f} '
            
            logger.info(message)
            
    @abstractmethod
    def on_train_batch_end(self, step: int, batch: Features):
        """ """
        
    @abstractmethod
    def on_test_batch_end(self, step: int, batch: Features):
        """ """


@dataclass
class LossHistory(History):
    """ """
    name: str = 'Loss'
    metric: Callable[[], tf.metrics.Metric] = None
    training_step: Callback = None
    verbose: bool = False
    
    def on_train_batch_end(self, step: int, batch: Features):
        train_step_outputs = self.training_step.train_step_output
        loss = train_step_outputs.loss
        self.train_metric.update_state(loss)
        '''
        if self.verbose:
            message = f'Train  ' \
                      f'Epoch: {self.epoch+1:4d} ' \
                      f'Step: {step+1:4d} ' \
                      f'Loss:     {self.train_metric.result():1.8f}'
            logger.info(message)
        '''
        
    def on_test_batch_end(self, step: int, batch: Features):
        test_step_outputs = self.training_step.test_step_output
        loss = test_step_outputs.loss
        self.test_metric.update_state(loss)
        '''
        if self.verbose:
            message = f'Test   ' \
                      f'Epoch: {self.epoch+1:4d} ' \
                      f'Step: {step+1:4d} ' \
                      f'Loss:     {self.test_metric.result():1.8f} '
            logger.info(message)
        '''   

        
@dataclass
class AccuracyHistory(History):
    """ """
    name: str = 'Accuracy'
    metric: Callable[[], tf.metrics.Metric] = None
    training_step: Callback = None
    verbose: bool = False
    
    def on_train_batch_end(self, step: int, batch: Features):
        train_step_outputs = self.training_step.train_step_output 
        y_pred = train_step_outputs.logits
        y_true = batch.label
        self.train_metric.update_state(y_true, y_pred)
        
    def on_test_batch_end(self, step: int, batch: Features):
        test_step_outputs = self.training_step.test_step_output 
        y_pred = test_step_outputs.logits
        y_true = batch.label
        self.test_metric.update_state(y_true, y_pred)
        
        
@dataclass
class F1ScoreHistory(History):
    name: str = 'F1-score'
    metric: Callable[[], tf.metrics.Metric] = None
    mode: str = None
    training_step: Callback = None
    verbose: bool = False
    
    def on_train_batch_end(self, step: int, batch: Features):
        train_step_outputs = self.training_step.train_step_output
        logits = train_step_outputs.logits
        
        if self.mode not in ['multiclass', 'multilabel']:
            raise ValueError
            
        if self.mode == 'multiclass':
            y_pred = tf.nn.softmax(logits)
            y_true = batch.label
        elif self.mode == 'multilabel':
            y_pred = tf.math.sigmoid(logits)
            y_true = batch.label
            
        self.train_metric.update_state(y_true, y_pred)  
        
    def on_test_batch_end(self, step: int, batch: Features):
        test_step_outputs = self.training_step.test_step_output
        logits = test_step_outputs.logits
        
        if self.mode not in ['multiclass', 'multilabel']:
            raise ValueError
            
        if self.mode == 'multiclass':
            y_pred = tf.nn.softmax(logits)
            y_true = batch.label
        elif self.mode == 'multilabel':
            y_pred = tf.math.sigmoid(logits)
            y_true = batch.label
            
        self.test_metric.update_state(y_true, y_pred)
        
        
@dataclass
class FBetaScoreHistory(History):
    name: str = 'FBeta-score'
    metric: Callable[[], tf.metrics.Metric] = None
    mode: str = None
    training_step: Callback = None
    verbose: bool = False
    
    def on_train_batch_end(self, step: int, batch: Features):
        train_step_outputs = self.training_step.train_step_output
        logits = train_step_outputs.logits
        
        if self.mode not in ['multiclass', 'multilabel']:
            raise ValueError
            
        if self.mode == 'multiclass':
            y_pred = tf.nn.softmax(logits)
            y_true = batch.label
        elif self.mode == 'multilabel':
            y_pred = tf.sigmoid(logits)
            y_true = batch.label
            
        self.train_metric.update_state(y_true, y_pred)
        
    def on_test_batch_end(self, step: int, batch: Features):
        test_step_outputs = self.training_step.test_step_output
        logits = test_step_outputs.logits
        
        if self.mode not in ['multiclass', 'multilabel']:
            raise ValueError
            
        if self.mode == 'multiclass':
            y_pred = tf.nn.softmax(logits)
            y_true = batch.label
        elif self.mode == 'multilabel':
            y_pred = tf.nn.sigmoid(logits)
            y_true = batch.label
            
        self.test_metric.update_state(y_true, y_pred)
        
        
@dataclass
class PrecisionHistory(History):
    name: str = 'Precision'    
    metric: Callable[[], tf.metrics.Metric] = None
    mode: str = None
    training_step: Callback = None
    verbose: bool = False
    
    def on_train_batch_end(self, step: int, batch: Features):
        train_step_outputs = self.training_step.train_step_output
        logits = train_step_outputs.logits
        
        if self.mode not in ['multiclass', 'multilabel']:
            raise ValueError
            
        if self.mode == 'multiclass':
            y_pred = tf.nn.softmax(logits)
            y_true = batch.label
        elif self.mode == 'multilabel':
            y_pred = tf.nn.sigmoid(logits)
            y_true = batch.label
            
        self.train_metric.update_state(y_true, y_pred)
        
    def on_test_batch_end(self, step: int, batch: Features):
        test_step_outputs = self.training_step.test_step_output
        logits = test_step_outputs.logits
        
        if self.mode not in ['multiclass', 'multilabel']:
            raise ValueError
            
        if self.mode == 'multiclass':
            y_pred = tf.nn.softmax(logits)
            y_true = batch.label
        elif self.mode == 'multilabel':
            y_pred = tf.nn.sigmoid(logits)
            y_true = batch.label
            
        self.test_metric.update_state(y_true, y_pred)
        
        
@dataclass
class RecallHistory(History):
    name: str = 'Recall'    
    metric: Callable[[], tf.metrics.Metric] = None
    mode: str = None
    training_step: Callback = None
    verbose: bool = False
    
    def on_train_batch_end(self, step: int, batch: Features):
        train_step_outputs = self.training_step.train_step_output
        logits = train_step_outputs.logits
        
        if self.mode not in ['multiclass', 'multilabel']:
            raise ValueError
            
        if self.mode == 'multiclass':
            y_pred = tf.nn.softmax(logits)
            y_true = batch.label
        elif self.mode == 'multilabel':
            y_pred = tf.nn.sigmoid(logits)
            y_true = batch.label
            
        self.train_metric.update_state(y_true, y_pred)
        
    def on_test_batch_end(self, step: int, batch: Features):
        test_step_outputs = self.training_step.test_step_output
        logits = test_step_outputs.logits
        
        if self.mode not in ['multiclass', 'multilabel']:
            raise ValueError
            
        if self.mode == 'multiclass':
            y_pred = tf.nn.softmax(logits)
            y_true = batch.label
        elif self.mode == 'multilabel':
            y_pred = tf.nn.sigmoid(logits)
            y_true = batch.label
            
        self.test_metric.update_state(y_true, y_pred)
        
        
@dataclass
class ModelCheckpoint(Callback):
    """ Callback that save the model weights and configuration 
    after each epoch and to keep track of the best model """
    
    name: str
    model: transformers.TFPreTrainedModel = None
    history: History = None
    direction: str = 'minimize'
    best_result: float = np.inf
    best_model_dir: str = ''
    checkpoints_dir: str = 'checkpoints'
    verbose: bool = False

    def __post_init__(self):
        """ Create the directory for saving checkpoints. """
        if self.direction not in ['minimize', 'maximize']:
            raise ValueError
        if self.direction == 'maximize':
            self.best_result = 0

        if not os.path.isdir(self.checkpoints_dir):
            abs_path = os.path.abspath(self.checkpoints_dir)
            text = f'Make a checkpoint directory: {abs_path}'
            logger.info(text)
            os.makedirs(self.checkpoints_dir)

    def on_epoch_end(self, epoch: int):
        """ Save the model checkpoint if the performance improves """
        result = self.history.test[epoch]
        diff = self.best_result - result
        is_better = diff > 0 if self.direction == 'minimize' else diff < 0
        if is_better:
            name = f'epoch-{epoch:02d}-{result:.2f}'
            model_dir = os.path.join(self.checkpoints_dir, name)
            os.mkdir(model_dir)
            self.model.config.architectures = self.model.__class__.__name__
            self.model.save_pretrained(model_dir)
            self.best_result = result
            self.best_model_dir = model_dir
            
            if self.verbose:
                message = f'New model checkpoint saved.'
                logger.info(message)
                
                
@dataclass
class CSVLogger(Callback):
    """ Callback that store the logs of the performance metrics
    into a csv file. """
    
    name: str
    logs_dir: str = None
    level: int = 20
    msg_format: str = '%(asctime)s [%(levelname)-5s] %(message)s'

    def __post_init__(self):
        root_logger = logging.getLogger('__name__')
        root_logger.setLevel(self.level)
        root_logger.propagate = False
        formatter = logging.Formatter(self.msg_format, 
                                      datefmt='%Y-%m-%d %H:%M:%S')
        console = logging.StreamHandler()
        console.setFormatter(formatter)
        
        root_logger.addHandler(console)
        if self.logs_dir:
            logs_dir = os.path.abspath(self.logs_dir)
            os.makedirs(logs_dir, exist_ok=True)
            file_path = os.path.join(logs_dir, 'experiment.log')
            file_handler = logging.FileHandler(file_path, mode='w')
            file_handler.setFormatter(formatter)
            root_logger.addHandler(file_handler)
            
            
@dataclass
class EarlyStopping(Callback):
    """ Callback that stop training the model once a certain metric
    is no longer improving. The metric used is typically the 
    validation loss, but it can also be a metric such as accuracy
    or precision """
    
    name: str = 'EarlyStopping'
    history: History = None
    patience: int = 3
    min_delta: float = 0.01
    best_result: float = np.inf
    current_patience: int = 0
    direction: str = 'maximize'

    def __post_init__(self):
        if self.direction not in ['minimize', 'maximize']:
            raise ValueError
        if self.direction == 'maximize':
            self.best_result = 0

    def on_epoch_end(self, epoch: int):
        """ Check if the model is improving after each epoch
        and stop the training process if no improvement after
        a certain amount of iteration."""
        result = self.history.test[epoch]
        diff = self.best_result - result
        is_better = diff > self.min_delta if self.direction == 'minimize' \
            else diff < self.min_delta * -1
        if is_better:
            self.best_result = result
            self.current_patience = 0
            return
        self.current_patience += 1
        if self.current_patience >= self.patience:
            raise Exception('Early stopping of the training')


@dataclass
class TimedStopping(Callback):
    """ Callbacks that stop training the model when a 
    specified amount of time has passed.
    
    Args:
        time_limit: maximum amount of time in seconds 
             before stopping.
        stopped_epoch: the epoch number where the training
              stopped.
    """

    name: str = 'TimedStopping'
    time_limit: int = 86000
    stopped_epoch: int = None

    def get_config(self):
        config = {"seconds": self.time_limit}
        base_config = super().get_config()
        return {**base_config, **config}
    
    def on_train_begin(self):
        self.stopping_time = time.time() + self.time_limit
        
    def on_epoch_end(self, epoch: int):
        if time.time() >= self.stopping_time:
            self.stopped_epoch = epoch
            formatted_time = datetime.timedelta(seconds=self.time_limit)
            message = "Timed stopping at epoch {} after training for {}".format(
                self.stopped_epoch + 1, formatted_time
            )
            raise Exception(message)

    def on_train_end(self):
        if self.stopped_epoch is not None:
            formatted_time = datetime.timedelta(seconds=self.time_limit)
            message = "Timed stopping at epoch {} after training for {}".format(
                self.stopped_epoch + 1, formatted_time
            )
            logger.info(message)


@dataclass
class ProgbarLogger(Callback):
    """ Callbacks that display a progress bar with performance metrics
    to monitor the progress of a model's training in the console """ 
    
    name: str = 'ProgbarLogger'
    history: History = None
    epoch: int = field(default_factory=int)
    num_samples: int = None
    batch_size: int = None
    show_overall_progress: bool = False
    show_epoch_progress: bool = True
    update_per_second: int = 10
    is_training = False

    def __post_init__(self):
        self.tqdm = tqdm
        self.num_steps: int = int(self.num_samples/self.batch_size)
        self.update_interval = 1/self.update_per_second
        self.last_update_time = time.time()
        self.bar_format: str = "Batch {n_fmt}/{total_fmt}{bar} " \
                               "ETA: {remaining}s - {desc}"
        self.metrics_format: str = "{name}: {value:0.4f}" 
        self.metrics_separator: str = " - "
        
    def _initialize_progbar(self, hook, epoch):
        self.num_samples_seen = 0
        self.steps_to_update = 0
        self.steps_so_far = 0
        
        if hook == "Train":
            current_epoch_description = "Epoch {epoch}".format(epoch=epoch+1)
            if self.show_epoch_progress:
                print(current_epoch_description)
                self.epoch_progress_tqdm=self.tqdm(
                    desc="Initialization",
                    total=self.num_steps,
                    position=0,
                    bar_format=self.bar_format,
                    leave=False,
                    dynamic_ncols=True,
                    unit="steps",
                )
        elif hook == "Test":
            if self.show_epoch_progress:
                self.epoch_progress_tqdm=self.tqdm(
                    desc="Initialization",
                    total=self.num_steps,
                    position=0,
                    bar_format=self.bar_format,
                    leave=False,
                    dynamic_ncols=True,
                    unit="steps",
                )

    def _update_progbar(self, hook):
        #self.num_samples_seen += self.batch_size
        self.steps_to_update += 1
        self.steps_so_far += 1

        if self.steps_so_far <= self.num_steps:  
            now = time.time()
            time_diff = now - self.last_update_time
            if self.show_epoch_progress and time_diff >= self.update_interval:
                
                desc = f"Step: {hook} - "
                for metric in self.history:
                    desc += f"{metric.name}: {metric.train_metric.result():0.4f} - "
                self.epoch_progress_tqdm.desc = desc
                self.epoch_progress_tqdm.update(self.steps_to_update)
                self.steps_to_update = 0
                self.last_update_time = now
                
    def _clean_up_progbar(self, hook):
        if self.show_epoch_progress:
            self.epoch_progress_tqdm.miniters = 0
            self.epoch_progress_tqdm.mininterval = 0
            self.epoch_progress_tqdm.update(
               self.num_steps - self.epoch_progress_tqdm.n
           )
            self.epoch_progress_tqdm.close()
                 
    def on_train_begin(self):
        self.is_training = True
        
    def on_train_end(self):
        self.is_training = False
        self._clean_up_progbar("Train")
        
    def on_test_begin(self):
        if not self.is_training:
            self._initialize_progbar("Test", epoch=None)
            
    def on_test_end(self):
        if not self.is_training:
            self._clean_up_progbar("Test")
            
    def on_epoch_begin(self, epoch: int):
        self.epoch = epoch
        self._initialize_progbar("Train", epoch)
    
    def on_epoch_end(self, epoch: int):
        self._clean_up_progbar("Train")
        if self.show_overall_progress:
            self.overall_progress_tqdm.update(1)
        self.epoch_progress_tqdm.close()

    def on_train_batch_end(self, step: int, batch: Features):
        self._update_progbar("Train")
        
    def on_test_batch_end(self, step: int, batch: Features):
        self._update_progbar("Test")
        


from typing import Any
from typing import Callable
from typing import Iterable
from typing import List
from typing import Optional
from typing import Tuple

from itertools import chain
from collections import namedtuple
from collections import Counter

import math
import numpy as np
import tensorflow as tf

from transformers import PretrainedConfig

from sklearn.preprocessing import MultiLabelBinarizer

from nlpipes.data.data_types import Example


def chunk() -> Callable[[str], List[str]]:
    """  Chunk text into shorter sequences. """  
    
    def chunk_into_sentences(texts: str) -> List[str]:
        """ Chunk text with regular-expressions """
        on_regex = '(?<=[^A-Z].[.?]) +(?=[A-Z])'
        sequences = [re.split(on_regex, text) for text in texts]
        sequences = [sequence for sublist in sequences for sequence in sublist]
        return sequences
    
    def chunk_into_whatever(texts: str) -> List[str]:
        raise NotImplementedError
    
    chunk_fn = chunk_into_sentences

    return chunk_fn


def create_examples(texts: List[str]) -> List[Example]:
    """ Serialize each input raw data into unlabelled examples. """
    examples = [Example(text) for text in texts]
    return examples


def create_monolabel_examples(
    texts: List[str],
    labels: Optional[List[str]],
    config: PretrainedConfig,
   ) -> List[Example]:
    """ Serialize each input raw data into single label examples. """
    labels = [config.label2id[label] for label in labels] 
    examples = [
        Example(text, label=label)
        for text, label in zip(texts, labels)
    ]        
    return examples


def create_multilabel_examples(
    texts: List[str],
    labels: Optional[List[List[str]]],
    config: PretrainedConfig,
   ) -> List[Example]:
    """ Serialize each input raw data into multiple label examples. """
    labels = [
        [config.label2id[label] for label in sub_labels]
        for sub_labels in labels
    ]
    examples = [
        Example(text, label=label)
        for text, label in zip(texts, labels)
    ]
    return examples

        
def create_classlabel_examples(
    texts: List[str],
    labels: Optional[List[List[List[str]]]],
    config: PretrainedConfig,
   ) -> List[Example]:
    """ Serialize each input raw data into class label examples. """
    
    def generate_classlabel_example(
        text: str,
        labels: List[Any],
        config: PretrainedConfig,
    ):
        label2id = {
            'negative': 0,
            'neutral': 1,
            'positive': 2,
        }

        aspects = [label[0] for label in labels]
        polarities = [label[1] for label in labels]

        for i, aspect in enumerate(aspects):
            aspect = aspect.lower()
            sentiment = polarities[i]
            label = label2id[sentiment]
            yield Example(
                text=text, 
                aspect=aspect, 
                label=label,
            )
    
    data = namedtuple('data', 'x y')
    data_raw = data(texts, labels)
    
    examples = [
        example
        for x, y in zip(*data_raw)
        for example in generate_classlabel_example(x, y, config)
    ]
    return examples
    

def split_examples(
    examples: List[Example],
    train_frac: float = None,
    test_frac: float = None, 
    shuffle: bool = True,
    random_seed: int = 42,
   ) -> Tuple[Example]:
    """ split the input example into a training and a
    testing set. """
    n_examples = len(examples)
    n_train = int(np.ceil(n_examples*train_frac))
    n_test = int(np.ceil(n_examples*test_frac))
    n_total = n_train + n_test
    
    if shuffle:
        random_state = np.random.RandomState(random_seed)
        order = random_state.permutation(n_examples)
        train_idx = order[:n_train]
        test_idx = order[n_train:n_total]
    else:
        train_idx = np.arange(n_train)
        test_idx = np.arange(n_train, n_total)
    
    train_examples = [examples[idx] for idx in train_idx]
    test_examples = [examples[idx] for idx in test_idx]
    return train_examples, test_examples


def generate_batches(
    X: List[str], 
    Y: List[Any], 
    batch_size: int,
) -> Iterable[List[Any]]:
    """ Yield a batch of examples from the input raw data """
    X_batch = list()
    Y_batch = list()
    
    for x, y in zip(X,Y):
        X_batch.append(x)
        Y_batch.append(y)
        if len(X_batch) < batch_size:
            continue
        yield X_batch, Y_batch    
        X_batch = list()
        Y_batch = list()

            
def get_labels_depth(labels: List[Any]) -> int:
    """ Get the depth level of a potentially nested list of labels """
    level = 1
    if isinstance(labels, list):
        return level + max(get_labels_depth(label) for label in labels)
    else:
        return 0
    
    
def make_alignment(
    labels: List[Tuple], 
    alignements: List[int],
) -> List[Any]:
    """ reformat the predicted labels to keep alignment
    with the input text."""
    n_labels = len(set(alignements))
    aligned_labels = [[] for idx in range(n_labels)]
    
    for idx, position in enumerate(alignements):
        aligned_labels[position].append(labels[idx])
    
    return aligned_labels
    

def convert_onehot_to_ids(
    one_hot: tf.Tensor
) -> tf.Tensor:
    """ """
    mask = tf.dtypes.cast(one_hot, tf.bool)
    s = tf.shape(mask)
    r = tf.reshape(
            tf.range(s[-1]),
            tf.concat([
                tf.ones(
                    tf.rank(one_hot) - 1, 
                    tf.int32),
                [-1]
            ],
            axis=0)
    )
    r = tf.tile(r, tf.concat([s[:-1], [1]], axis=0))
    ragged_tensor = tf.ragged.boolean_mask(r, mask)
    return ragged_tensor


def convert_ids_to_onehot(ids, depth, dtype):
    """ """
    one_hot_tensor = tf.convert_to_tensor(
        MultiLabelBinarizer(
            classes=range(depth)
        ).fit_transform(ids),
        dtype=dtype
    )
    return one_hot_tensor


def get_class_names(Y: list, classes=None):
    """ """
    if classes is None:
        classes = set()
    for y in Y:
        if isinstance(y, list):
            get_classes(y, classes)
        else:
            classes.add(y)
    class_names = list(sorted(classes))    
    return class_names


def get_class_weights(Y: list, mu: float=0.5):
    """ """
    Y_flatten = [y for sub_Y in Y for y in sub_Y]
    counter = Counter(sorted(Y_flatten))
    class_names = [k for k in counter.keys()]
    class_freqs = [v for v in counter.values()]
    total = sum(class_freqs)
    
    class_weights = []
    for key in counter.keys():
        weight = math.log(mu*total/float(counter[key]))
        weight = weight if weight > 1.0 else 1.0
        class_weights.append(weight)
    
    return class_names, class_weights


from dataclasses import dataclass
from itertools import chain
from typing import Iterable
from typing import List
from typing import Optional

import numpy as np

from nlpipes.data.data_types import Example


@dataclass
class TrainTestSplitter:
    """ Splits input data into training/test set """
        
    train_frac: float = None
    test_frac: float = None
    shufffle: bool = True

    def __call__(self, examples: List[Example]
                ) -> Iterable:
        """  """
        
        n_examples = len(examples)
        n_train = int(np.ceil(n_examples*self.train_frac))
        n_test = int(np.ceil(n_examples*self.test_frac))
        n_samples = n_train + n_test
        
        if self.shufffle:
            order = np.random.permutation(n_examples)
            train_indices = order[n_train:]
            test_indices = order[:n_test]
        else:
            train_indices = np.arange(n_train)
            test_indices = np.arange(n_train, n_samples)
        
        train_examples = self._indexing(examples, train_indices)
        test_examples = self._indexing(examples, test_indices) 
            
        return train_examples, test_examples

    def _indexing(self, examples, indices):
        """  """
        
        if hasattr(examples, 'shape'):
            return examples[indices]
        
        return [examples[idx] for idx in indices]
    


import os
import logging
import importlib
import inspect

from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
)

import tensorflow as tf
import transformers

from tensorflow.python.client import device_lib

from transformers import (
    TFPreTrainedModel,
    TFAutoModelForMaskedLM,
    TFAutoModelForSequenceClassification,
    TFBertForSequenceClassification,
    BertTokenizerFast,
    AutoModel,
    AutoTokenizer,
    AutoConfig, 
)

from nlpipes.pipelines.base_pipeline import Pipeline
from nlpipes.pipelines.language_modeling_pipelines import PipelineForMaskedLanguageModeling
from nlpipes.pipelines.mono_label_classification_pipelines import PipelineForMonoLabelSequenceClassification
from nlpipes.pipelines.multi_label_classification_pipelines import PipelineForMultiLabelSequenceClassification
from nlpipes.pipelines.class_label_classification_pipelines import PipelineForClassLabelSequenceClassification
from nlpipes.models.classification_models import TFAdaptedBertForSequenceClassification
from nlpipes.models.classification_models import AdaptedBertForSequenceClassification
from nlpipes.configurations.bert_config import AdaptedBertConfig
from nlpipes.data.data_types import Task
from nlpipes.data.data_types import LabelsMapping

os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
tf.get_logger().setLevel('ERROR')

logger = logging.getLogger('__name__')


SUPPORTED_TASKS = {
  "language-modeling": {
      "pipeline": PipelineForMaskedLanguageModeling,
      "models": {
          "default": TFAutoModelForMaskedLM,
          "adapters": None,
      },
      "tokenizers": {
          "default": AutoTokenizer,
          "adapters": AutoTokenizer,
      },
      "configs": {
          "default": AutoConfig,
          "adapters": AdaptedBertConfig,
      },
  },
  "mono-label-classification": {
      "pipeline": PipelineForMonoLabelSequenceClassification,
      "models": {
          "default": TFAutoModelForSequenceClassification,
          "adapters": TFAdaptedBertForSequenceClassification,
      },
      "tokenizers": {
          "default": AutoTokenizer,
          "adapters": AutoTokenizer,
      },
      "configs": {
          "default": AutoConfig,
          "adapters": AdaptedBertConfig,
      },
  },
  "multi-label-classification": {
      "pipeline": PipelineForMultiLabelSequenceClassification,
      "models": {
          "default": TFAutoModelForSequenceClassification,
          "adapters": TFAdaptedBertForSequenceClassification,
      },
      "tokenizers": {
          "default": AutoTokenizer,
          "adapters": AutoTokenizer,
      },
      "configs": {
          "default": AutoConfig,
          "adapters": AdaptedBertConfig,
      },
  },
  "class-label-classification": {
      "pipeline": PipelineForClassLabelSequenceClassification,
      "models": {
          "default": TFAutoModelForSequenceClassification,
          "adapters": TFAdaptedBertForSequenceClassification,
      },
      "tokenizers": {
          "default": AutoTokenizer,
          "adapters": AutoTokenizer,
      },
      "configs": {
          "default": AutoConfig,
          "adapters": AdaptedBertConfig,
      },
  },
}


def Model(
    name_or_path: TFPreTrainedModel = None,
    task: str = None,
    all_labels: Optional[List] = None,
    from_pt:bool = False,
    **model_kwargs
) -> Pipeline:

    """ The `Model` offers an abstraction to run several text 
    classification tasks. Which task the model should learn 
    is determined automatically based on the user-supplied
    target task.

    Args
    ----------
    task(str):
          A task name that determines which task the model 
          should learn e.g., `single_label` classification`.
            
    model(str):
          A path or name of the model that should be used.
          It can be:
            - The path of a model available in a local repo.
            - The name of a model available in the Huggingface
              model repo.
    
    all_labels (List[str]): 
          A list of all labels. Needed as some labels may be 
          not present in the dataset.
    """

    try:
        pipeline = get_pipeline_from_task(
            task=task, 
            name_or_path=name_or_path, 
            all_labels=all_labels,
            customization_tag='#',
            from_pt=from_pt,
        )

        return pipeline

    except EnvironmentError as error:
        message = "The pipeline could not be loaded."
        logger.error(message)
        raise error


def get_pipeline_from_task(
    task:str, 
    name_or_path:str,
    all_labels:List[str],
    customization_tag:str,
    from_pt:bool,
) -> Pipeline:
    """ Create the pipeline given a downstream task. """
    
    if task:
        TASK = SUPPORTED_TASKS[task]
        PIPELINE = TASK["pipeline"]
    
        if customization_tag in name_or_path:
            customization = name_or_path.split(customization_tag, 1)[1]
            MODEL = TASK["models"][customization]
            TOKENIZER = TASK["tokenizers"][customization]
            CONFIG = TASK["configs"][customization]
            name_or_path = name_or_path.replace(
                customization_tag+customization, ""
            )
        else:
            MODEL = TASK["models"]["default"]
            TOKENIZER = TASK["tokenizers"]["default"]
            CONFIG = TASK["configs"]["default"]

        config = CONFIG.from_pretrained(name_or_path)
        
        if all_labels:
            all_labels = sorted(all_labels)
            config.id2label=LabelsMapping(all_labels).id2label()
            config.label2id=LabelsMapping(all_labels).label2id()
            config.num_labels=LabelsMapping(all_labels).num_labels()
            
        if not config.finetuning_task:
            config.finetuning_task=task

        tokenizer = TOKENIZER.from_pretrained(name_or_path)
        model = MODEL.from_pretrained(
            name_or_path, 
            config=config, 
            from_pt=from_pt,
        )
        
        pipeline = PIPELINE(
            model=model,
            tokenizer=tokenizer,
            config=config,
        )
    
    else:
        config = AutoConfig.from_pretrained(name_or_path)
        PIPELINE = SUPPORTED_TASKS[config.finetuning_task]['pipeline']
        
        MODEL = get_model_from_config(config)
        
        model = MODEL.from_pretrained(name_or_path)
        tokenizer = AutoTokenizer.from_pretrained(name_or_path)
        
        pipeline = PIPELINE(
            model=model,
            tokenizer=tokenizer,
            config=config,
        )


    return pipeline


def get_model_from_config(config):
    """ Get the model class from the model configuration file.
    Note : we use a temporary dirty workaround to bypass an issue in 
    the model architecture name stored in the saved config """
    
    model_name = config.architectures[0]
    models_module_name = 'nlpipes.models.classification_models'
    models_module = importlib.import_module(models_module_name)
    available_models = [
        model[0] for model in inspect.getmembers(models_module)
    ]
    
    if model_name in available_models:
        model_class = getattr(models_module, model_name)
    else:
        model_class = getattr(transformers, 'TF' + model_name)
    return model_class


def get_available_tasks() -> List[str]:
    """ Get a list of all available tasks."""
    return list(SUPPORTED_TASKS.keys())


def get_available_gpus():
    local_device_protos = device_lib.list_local_devices()
    return [
        device.name for device in local_device_protos 
        if device.device_type == 'GPU'
    ]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nlpipes',
 'nlpipes..ipynb_checkpoints',
 'nlpipes.callbacks',
 'nlpipes.callbacks..ipynb_checkpoints',
 'nlpipes.configurations',
 'nlpipes.configurations..ipynb_checkpoints',
 'nlpipes.data',
 'nlpipes.data..ipynb_checkpoints',
 'nlpipes.layers',
 'nlpipes.layers..ipynb_checkpoints',
 'nlpipes.losses',
 'nlpipes.losses..ipynb_checkpoints',
 'nlpipes.metrics',
 'nlpipes.metrics..ipynb_checkpoints',
 'nlpipes.models',
 'nlpipes.models..ipynb_checkpoints',
 'nlpipes.optimization',
 'nlpipes.optimization..ipynb_checkpoints',
 'nlpipes.pipelines',
 'nlpipes.pipelines..ipynb_checkpoints',
 'nlpipes.trainers',
 'nlpipes.trainers..ipynb_checkpoints']

package_data = \
{'': ['*']}

install_requires = \
['ftfy>=6.1.1,<7.0.0',
 'numpy>=1.21.0,<2.0.0',
 'pandas>=1.3.5,<2.0.0',
 'scipy>=1.7.3,<2.0.0',
 'tensorflow>=2.11.0,<3.0.0',
 'tokenizers>=0.13.0,<0.14.0',
 'tqdm>=4.64.0,<5.0.0',
 'transformers>=4.24.0,<5.0.0']

setup_kwargs = {
    'name': 'nlpipes',
    'version': '0.1.17',
    'description': 'Text Classification with Transformers',
    'long_description': '<!-- PROJECT NAME -->\n<div align="center">\n   <img src="https://ik.imagekit.io/m0ci8dgk4/nlpipes_logo_eSBhzDKCZ.png?updatedAt=1679840445991" alt="nlpipes_logo" title="nlpipes logo">\n  <h2>Text Classification with Transformers</h2>\n</div>\n\n<div align="center">\n    <a href="https://opensource.org/licenses/Apache-2.0">\n       <img alt="Licence" src="https://img.shields.io/badge/License-Apache_2.0-blue.svg">\n    </a>\n     <a href="https://pypi.org/project/nlpipes/">\n       <img alt="PyPi Version" src="https://img.shields.io/pypi/pyversions/nlpipes">\n    </a> \n    <a href="https://pypi.org/project/nlpipes/">\n        <img alt="PyPi Package Version" src="https://img.shields.io/pypi/v/nlpipes">\n    </a>\n    <!--\n    <a href="https://pepy.tech/project/nlpipes/">\n        <img alt="PyPi Downloads" src="https://static.pepy.tech/badge/nlpipes/month">\n    </a>\n    -->\n</div>\n\n<div align="center">\n    <a href=""><strong>Documentation</strong></a>\n    • <a href=""><strong>References</strong></a>\n</div>\n\n\n<div align="center">\n  <img src="https://ik.imagekit.io/m0ci8dgk4/nlpipes_screenshot_q_FDeS4Js.png?updatedAt=1686231419316"\n       alt="nlpipes_screenshot" \n       title="nlpipes screenshot">\n</div>\n\n\n## Overview\n\n`NLPipes` is for people unfamiliar with Transformers who want an end to end solution to solve practical text classification problems, including:\n\n* **Single-label classification**: A typical use case is sentiment detection where one want to detect the overall sentiment polarity (e.g., positive, neutral, negative) in a review.\n* **Multi-label classification**: A typical use case is aspect categories detection where one want to detect the multiple aspects mentionned in a review (e.g., product_quality, delivery_time, price, ...).\n* **Class-label classification**: A typical use case is aspect based sentiment analysis where one want to detect sentiment polarity associated to each aspect categories mentionned in a review (e.g., product_quality: neutral, delivery_time: negative, price: positive, ...).\n\n`NLPipes` expose a `Model` API that provide a unique and simple abstraction for all the tasks. \nThe library maintain a common usage pattern across models (train, evaluate, predict, save) with\nalso a clear and consistent data structure (python lists as inputs/outputs data).\n\n#### Built with\n`NLPipes` is built with TensorFlow and HuggingFace Transformers:\n* [TensorFlow](https://www.tensorflow.org/): An end-to-end open source deep learning framework\n* [Transformers](https://huggingface.co/transformers/): An general-purpose open-sources library for transformers-based architectures\n\n## Getting Started\n\n#### Installation\n1. Create a virtual environment\n\n ```console\n python3 -m venv nlpipesenv\n source nlpipesenv/bin/activate\n ```\n\n2. Install the package\n\n ```console\n pip install nlpipes\n ```\n \n \n#### Tasks\n\nA model can be trained for a specific task by first loading a backbone model. The train command takes\nat minimum two parameters (X and Y), where X is a list of texts to train on and Y is the \ntraining target.\n\nThe training target expect different formats, depending on what task you want to solve:\n\n\n##### Single Label Classification:\nGive one label name for each sequence of text in `X`:\n\n ```console\n  model = Model("albert-base-v2",\n                task=\'single-label-classification\',\n                all_labels=["NEG", "NEU", "POS"],\n               )\n  \n  X = ["This was bad.", "This was great!"]\n  Y = ["NEG", "POS"]\n  \n  model.train(X, Y)\n ```\n\n##### Multiple Label Classification:\nGive a list of class names for each sequence of text in `X`:\n\n ```console\n  model = Model("albert-base-v2",\n                task=\'multi-label-classification\',\n                all_labels=all_labels,\n                )\n  \n  X = ["I want a refund!",\n       "The bill I got is not correct and I also have technical issues",\n       "All good"]\n  Y = [\n       ["billing"], \n       ["billing", "tech support"],\n       []\n      ]\n      \n  model.train(X, Y)\n ```\n \n ##### Aspect Based Classification:\nGive a list of lists of label lists (pairs) for each given text in `X`:\n\n ```console\n  model = Model("albert-base-v2",\n                task=\'class-label-classification\',\n                all_labels=["NEG", "NEU", "POS"],\n               )\n  \n  X = ["The room was nice.",\n       "The food was great, but the staff was unfriendly.",\n       "The room was horrible, but the waiters were welcoming"]\n  Y = [\n       [["room", "POS"],\n       [["food", "POS"], ["staff", "NEG"]],\n       [["room", "NEG"], ["staff", "POS"]],\n      ]\n      \n  model.train(X, Y)\n ```\n\n\n#### Examples\n\nHere are some examples on open datasets that show how to use `NLPipes` on different tasks:\n\nName|Notebook|Description|Task|Size|Memory|Speed| \n----|-----------|-----|---------|---------|---------|---------|\nGooglePlay Sentiment Detection|Available|Train a model to detect the sentiment polarity from the GooglePlay store |Single label classification|  |  |  \nStackOverflow tags Detection|Available|Train a model to detect tags from the StackOverFlow questions |Multiple label classification|  |  |\nAmazon Aspect Based Sentiment Detection|Available|Train a model to detect the aspect based sentiment polarity on Laptops Amazon reviews |Class label classification|  |  | \n\n\n## Notices\n- `NLPipes` is still in its early stage. The library comes with no warranty and future releases could bring substantial API and behavior changes.\n\n',
    'author': 'Ayhan UYANIK',
    'author_email': 'ayhan.uyanik@renault.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import logging

from typing import Iterable
from typing import List

from dataclasses import dataclass

import tensorflow as tf

from nlpipes.callbacks.callbacks import Callback
from nlpipes.callbacks.callbacks import CallbackList
from nlpipes.data.data_types import Features


logger = logging.getLogger('__name__')

@dataclass
class Trainer():
    
    """ The trainer is written from scratch. It consists in 
    a custom training loop that trigger a set of `Callback`
    functions at given stages of the training procedure, 
    such as when:
     - the training/testing starts.
     - the training/testing ends.
     - a train/test epoch starts
     - a train/test epoch ends.
     - a train/test batch starts.
     - a train/test batch ends.
    """
    
    callbacks: List[Callback]
    
    def fit(
        self,
        train_stream: Iterable[Features],
        test_stream: Iterable[Features],
        epochs: int,
    ):
        
        callbacks = CallbackList(self.callbacks)
        
        callbacks.on_train_begin()

        for epoch in range(epochs):
            
            callbacks.on_epoch_begin(epoch)
            for step, batch in enumerate(train_stream):
                callbacks.on_train_batch_begin(step, batch)
                callbacks.on_train_batch_end(step, batch)

            if test_stream:
                callbacks.on_test_begin()
                for step, batch in enumerate(test_stream):
                    callbacks.on_test_batch_begin(step, batch)
                    callbacks.on_test_batch_end(step, batch)

                callbacks.on_test_end()

            callbacks.on_epoch_end(epoch)
            
        callbacks.on_train_end()
        


from dataclasses import dataclass
from typing import List
from collections import Counter

import tensorflow as tf
import tensorflow_addons as tfa

import numpy as np
import math


@dataclass
class SCELoss():
    """ Computes the softmax cross-entropy """

    def __call__(self, labels: tf.Tensor, logits: tf.Tensor):
        softmax = tf.nn.softmax_cross_entropy_with_logits
        return softmax(labels, logits, axis=-1)
    
    
@dataclass
class SFCELoss():
    """ Computes the focal sigmoid cross-entropy """
    alpha: float=0.25
    gamma: float=2.0

    def __call__(self, labels: tf.Tensor, logits: tf.Tensor):
        sigmoid_focal = tfa.losses.sigmoid_focal_crossentropy
        return sigmoid_focal(
            labels, logits, 
            alpha=tf.constant(self.alpha), 
            gamma=tf.constant(self.gamma), 
            from_logits=True,
        )

    
@dataclass
class WCELoss():
    """ Computes the weighted sigmoid cross-entropy """
    class_weights: List[float]
    
    def __call__(self, labels: tf.Tensor, logits: tf.Tensor):
        weighted_sigmoid = tf.nn.weighted_cross_entropy_with_logits
        
        if self.class_weights:
            batch_size, n_classes = logits.get_shape()
            class_weights = tf.constant(self.class_weights)
            class_weights = tf.expand_dims(class_weights, axis=0)
            class_weights = tf.tile(class_weights, [batch_size, 1])
        else:
            class_weights = tf.constant(1.0)
                                        
        return weighted_sigmoid(labels, logits, class_weights)
    